import os
import sys
from time import time as current_time
from typing import Any, Callable, cast

from senpwai.utils.scraper import AnimeMetadata, Download, sanitise_title
from senpwai.utils.static import ROOT_DIRECTORY

from senpwai.scrapers import gogo, pahe

PAHE = pahe.PAHE
GOGO = gogo.GOGO

DEFAULT_ANIME_TITLE = "Senyuu"
DEFAULT_QUALITY = "360p"
DEFAULT_START_EPISODE = "1"
DEFAULT_END_EPISODE = "2"
DEFAULT_SUB_OR_DUB = "sub"
DEFAULT_SITE = "pahe"
DEFAULT_DOWNLOAD_FOLDER_PATH = os.path.join(ROOT_DIRECTORY, "..", ".test-downloads")
DEFAULT_VERBOSE = False
DEFAULT_SILENT = False
SILENT = DEFAULT_SILENT
COMMANDS = [
    "search",
    "dub_available",
    "metadata",
    "episode_page",
    "download_page",
    "download_size",
    "direct_links",
    "hls_links",
    "match_links",
    "segments_urls",
    "download",
    "all",
]


class FailedTest(Exception):
    def __init__(self, msg: str):
        super().__init__(msg)

def conditional_print(text: str):
    if not SILENT:
        print(text)


def test_start(name: str):
    conditional_print(f"Running: {name} Test")


def fail_test(
    name: str,
    expected: Any,
    got: Any,
    execution_time: float,
    test_variables: str | None = None,
):
    msg = f"{name} Test\nExpected: {expected}\nGot: {got}\nExecution time: {round(execution_time, 2)}s"
    if test_variables:
        msg = f"{msg}\nTest Variables:\n{test_variables}"
    raise FailedTest(msg)


def pass_test(name: str, execution_time: float):
    conditional_print(f"Passed: {name} Test\nTook: {round(execution_time, 2)}s\n")


def test_search(
    anime_title: str, site: str
) -> list[tuple[str, str]] | list[tuple[str, str, str]]:
    test_name = "Search"
    test_start(test_name)
    run_time_getter = get_run_time_later()
    if site == PAHE:
        results = pahe.search(anime_title)
    else:
        results = gogo.search(anime_title)
    rt = run_time_getter()
    if not results:
        fail_if_list_is_empty(results, test_name, "results", rt)
    pass_test(test_name, rt)
    test_name = "Parse search results"
    test_start(test_name)
    c = current_time()
    parsed_results: list[tuple[str, str]] | list[tuple[str, str, str]] = []
    if site == PAHE:
        parsed_results = [
            pahe.extract_anime_title_page_link_and_id(r)
            for r in cast(list[dict[str, str]], results)
        ]
    else:
        parsed_results = cast(list[tuple[str, str]], parsed_results)
        for title, page_link in cast(list[tuple[str, str]], results):
            parsed_results.append((title, page_link))
    run_timer = current_time() - c
    if not parsed_results:
        fail_test(test_name, "List of parsed results", "Empty list", run_timer)
    pass_test(test_name, run_timer)
    return parsed_results


def get_run_time_later() -> Callable[[], float]:
    c = current_time()
    return lambda: current_time() - c


def test_check_results_for_anime(
    anime_title: str,
    results: list[tuple[str, str]] | list[tuple[str, str, str]],
) -> tuple[str, str] | tuple[str, str, str] | None:
    test_name = f"Matching results to {anime_title}"
    test_start(test_name)
    run_time = get_run_time_later()
    query = sanitise_title(anime_title, True).lower()
    for r in results:
        title = r[0]
        if query == sanitise_title(title, True).lower():
            pass_test(test_name, run_time())
            return r
    fail_test(
        test_name,
        "A perfect match",
        "No match found",
        run_time(),
        f"Anime title was: {anime_title}\nSearch results were: {results}",
    )


def test_get_metadata(
    site: str, target_result: tuple[str, str] | tuple[str, str, str]
) -> tuple[AnimeMetadata, bytes]:
    test_name = "Get Metadata"
    test_start(test_name)
    run_time = get_run_time_later()
    page_content = b""
    if site == PAHE:
        target_result = cast(tuple[str, str, str], target_result)
        metadata = pahe.get_anime_metadata(target_result[2])
    else:
        page_content = gogo.get_anime_page_content(target_result[1])
        metadata = gogo.extract_anime_metadata(page_content)
    pass_test(test_name, run_time())
    return metadata, page_content


def test_get_episode_page_links(
    anime_id: str, anime_page_link: str, start_episode: int, end_episode: int
) -> list[str]:
    test_name = "Get Episode page links"
    test_start(test_name)
    run_time = get_run_time_later()
    test_variables = f"Anime ID: {anime_id}"
    anime_id = cast(str, anime_id)
    start_page_num, end_page_num, _, first_page = pahe.get_episode_pages_info(anime_page_link, start_episode, end_episode)
    episode_page_links = pahe.GetEpisodePageLinks().get_episode_page_links(
        start_episode, end_episode, start_page_num, end_page_num, first_page, anime_page_link, anime_id
    )
    rt = run_time()
    fail_if_list_is_empty(
        episode_page_links, test_name, "Episode page links", rt, test_variables
    )
    pass_test(test_name, run_time())
    return episode_page_links


def test_get_download_page_links(
    site: str,
    quality: str,
    sub_or_dub: str,
    eps_page_links: list[str],
    start_episode: int,
    end_episode: int,
    anime_id: int,
) -> tuple[list[str], list[str]]:
    test_name = "Get Download page links"
    test_start(test_name)
    run_time_getter = get_run_time_later()
    if site == PAHE:
        (
            pahewin_page,
            pahewin_info,
        ) = pahe.GetPahewinPageLinks().get_pahewin_page_links_and_info(
            eps_page_links
        )
        download_page_links, download_info = pahe.bind_sub_or_dub_to_link_info(
            sub_or_dub, pahewin_page, pahewin_info
        )
        fail_if_list_is_empty(
            download_page_links[0],
            test_name,
            f"Bound to {sub_or_dub} download page links",
            run_time_getter(),
            f"Episode page links were: {eps_page_links}\nPahewin page links were: {pahewin_page}",
        )
        download_page_links, download_info = pahe.bind_quality_to_link_info(
            quality, download_page_links, download_info
        )
        rt = run_time_getter()
        fail_if_list_is_empty(
            download_page_links,
            test_name,
            "Download page links",
            rt,
            f"Episode page links were: {eps_page_links}",
        )
    else:
        download_page_links = gogo.get_download_page_links(
            start_episode, end_episode, anime_id
        )
        download_info = []
        rt = run_time_getter()
        fail_if_list_is_empty(
            download_page_links,
            test_name,
            "Download page links",
            rt,
            f"Anime ID was: {anime_id}\nStart Episode was: {start_episode}\nEnd Episode was: {end_episode}",
        )
    pass_test(test_name, run_time_getter())
    return download_page_links, download_info


def fail_if_list_is_empty(
    array: list[Any],
    test_name: str,
    list_of: str,
    execution_time: float,
    test_variables: str | None = None,
):
    if not array:
        fail_test(
            test_name,
            f"List of {list_of}",
            "Empty list",
            execution_time,
            test_variables,
        )

def test_getting_direct_download_links(
    site: str, download_page_links: list[str], quality: str
) -> list[str]:
    test_name = "Get Direct download links"
    test_start(test_name)
    run_time_getter = get_run_time_later()
    ddls: list[str] = []
    if site == PAHE:
        ddls = pahe.GetDirectDownloadLinks().get_direct_download_links(
            download_page_links
        )
    else:
        ddls = gogo.GetDirectDownloadLinks().get_direct_download_links(
            download_page_links, quality
        )
    rt = run_time_getter()
    fail_if_list_is_empty(
        ddls,
        test_name,
        "direct download links",
        rt,
        f"Download page links were: {download_page_links}",
    )
    pass_test(test_name, rt)
    return ddls


def test_getting_hls_links(episode_page_links: list[str]) -> list[str]:
    test_name = "Get HLS links"
    test_start(test_name)
    runtime_getter = get_run_time_later()
    hls_links = gogo.GetHlsLinks().get_hls_links(episode_page_links)
    rt = runtime_getter()
    fail_if_list_is_empty(
        hls_links,
        test_name,
        "HLS links",
        rt,
        f"Episode page links were: {episode_page_links}",
    )
    pass_test(test_name, rt)
    return hls_links


def test_matching_quality_to_hls_links(hls_links: list[str], quality: str) -> list[str]:
    test_name = "Match Quality to HLS links"
    test_start(test_name)
    runtime_getter = get_run_time_later()
    hls_links = gogo.GetHlsMatchedQualityLinks().get_hls_matched_quality_links(
        hls_links, quality
    )
    rt = runtime_getter()
    fail_if_list_is_empty(
        hls_links,
        test_name,
        "Matched HLS links",
        rt,
        f"Original HLS links were: {hls_links}",
    )
    pass_test(test_name, rt)
    return hls_links


def test_getting_segments_urls(matched_hls_links: list[str]) -> list[list[str]]:
    test_name = "Get Segments URLs"
    test_start(test_name)
    runtime_getter = get_run_time_later()
    segs_urls = gogo.GetHlsSegmentsUrls().get_hls_segments_urls(matched_hls_links)
    rt = runtime_getter()
    fail_if_list_is_empty(
        segs_urls,
        test_name,
        "List containing List of Segment URLs",
        rt,
        f"Matched HLS links were: {matched_hls_links}",
    )
    fail_if_list_is_empty(
        segs_urls[0],
        test_name,
        "Segment URLs",
        rt,
        f"Matched HLS links were: {matched_hls_links}",
    )
    pass_test(test_name, runtime_getter())
    return segs_urls


def test_downloading(
    anime_title: str,
    ddls_or_segs_urls: list[str] | list[list[str]],
    is_hls_download: bool,
    start_eps: int,
    end_eps: int,
    download_folder: str,
):
    test_name = "Downloading"
    test_start(test_name)
    if not os.path.isdir(download_folder):
        os.makedirs(download_folder)
    runtime_getter = get_run_time_later()
    conditional_print(f"Folder: {download_folder}")
    for eps_no, ddl_or_seg_urls in zip(
        range(start_eps, end_eps + 1), ddls_or_segs_urls
    ):
        runtime_getter = get_run_time_later()
        eps_number = str(eps_no).zfill(2)
        eps_title = f"{anime_title} E{eps_number}"
        inner_test_name = f"Downloading {eps_title}"
        test_start(test_name)
        Download(
            ddl_or_seg_urls, eps_title, download_folder, is_hls_download=is_hls_download
        ).start_download()
        full_name = f"{eps_title}.mp4"
        rt = runtime_getter()
        downloaded_file = os.path.join(download_folder, full_name)
        if os.path.isfile(downloaded_file):
            pass_test(inner_test_name, rt)
            os.unlink(downloaded_file)
            continue
        test_variables = (
            f"HLS links were: {ddls_or_segs_urls}"
            if is_hls_download
            else f"DDLs were: {ddls_or_segs_urls}"
        )
        fail_test(
            inner_test_name,
            f"{full_name} file in {download_folder}",
            "Didn't find the file",
            rt,
            test_variables,
        )

    rt = runtime_getter()
    pass_test(test_name, rt)
    os.removedirs(download_folder)


class ArgParser:
    arg_site = ("--site", "-s")
    arg_verbose = ("--verbose", "-v")
    arg_title = ("--title", "-t")
    arg_quality = ("--quality", "-q")
    arg_sub_or_dub = ("--sub_or_dub", "-sd")
    arg_path = ("--path", "-p")
    arg_help = ("--help", "-h")
    arg_silent = ("--silent", "-s")

    def __init__(self, args: list[str]):
        self.passed_args = args[1:]
        global SILENT
        SILENT = self.arg_in_group_was_passed(self.arg_silent)
        if not self.arg_in_group_was_passed(COMMANDS):
            print("No valid Command specified")
            self.print_usage()
        if self.arg_in_group_was_passed(self.arg_help):
            self.print_usage()

        self.site = self.arg_value_finder(args, self.arg_site, DEFAULT_SITE)
        self.validate_arg_value(self.arg_site, (PAHE, GOGO), self.site)

        self.anime_title = self.arg_value_finder(
            args, self.arg_title, DEFAULT_ANIME_TITLE
        )
        self.anime_title = sanitise_title(self.anime_title)

        self.verbose = self.arg_in_group_was_passed(self.arg_verbose)
        self.start_eps, self.end_eps = self.parse_start_and_end_episode(args)

        self.quality = self.arg_value_finder(args, self.arg_quality, DEFAULT_QUALITY)
        self.validate_arg_value(
            self.arg_quality, ("360p", "480p", "720p", "1080p"), self.quality
        )

        self.sub_or_dub = self.arg_value_finder(
            args, self.arg_sub_or_dub, DEFAULT_SUB_OR_DUB
        )
        self.validate_arg_value(self.arg_sub_or_dub, ("sub", "dub"), self.sub_or_dub)

        self.path = self.arg_value_finder(
            args, self.arg_path, DEFAULT_DOWNLOAD_FOLDER_PATH
        )

    def print_usage(self):
        usage = """
        Usage: scrapers.test [COMMAND/TEST] [OPTIONS]

        Commands/Tests:
        search                  Test searching
        dub_available           Test dub availablity checking
        metadata                Test getting metadata
        episode_page            Test getting episode page links (pahe only)
        download_page           Test getting download page links
        download_size           Test extraction (pahe)/ getting (gogo) of total download size
        direct_links            Test getting direct download links
        hls_links               Test getting hls links
        match_links             Test matching hls links to user quality
        segments_urls           Test getting segments urls
        download                Test downloading (Implicitly performs all tests)
        all                     Perform all tests (alias to download). Only performs all tests for one site, defaults to pahe
        

        Options:
        --site, -s              Specify the site (i.e., pahe, gogo). Default: pahe
        --title, -t             Specify the anime title. Default: Senyuu
        --quality, -q           Specify the video quality (i.e., 360p, 480p, 720p, 1080p). Default: 360p
        --sub_or_dub, -sd       Specify sub or dub. Default: sub
        --path, -p              Specify the download folder path. Default: ./src/test-downloads
        --start_episode, -se    Specify the starting episode number. Default: 1
        --end_episode, -ee      Specify the ending episode number. Default: 2
        --verbose, -v           Enable verbose mode for more detailed explanations of test results
        --silent, s             Only output critical information
        --help, -h              Display this help message


        Example:
        scrapers.test --site pahe --title "Naruto" --quality 720p --sub_or_dub sub

        Note: Tests are hierarchial, if for example episode_page is the command/test to run, both search and metadata test will be performed cause they are required, in order to get episode_page links
        Note: Tests are site specific, the all command does not perform tests for each site it implicitly defaults to pahe
        """
        print(usage)
        sys.exit(0)

    def arg_in_group_was_passed(self, arg_group: tuple | list[str]):
        return any(arg in self.passed_args for arg in arg_group)

    def parse_start_and_end_episode(self, args: list[str]) -> tuple[int, int]:
        s = ("--start_episode", "-se")
        start_str = self.arg_value_finder(args, s, DEFAULT_START_EPISODE)
        e = ("--end_episode", "-ee")
        end_str = self.arg_value_finder(args, e, DEFAULT_END_EPISODE)

        def check_if_digit(x, v):
            return (
                self.invalid_usage(f"{x} MUST be a number") if not v.isdigit() else None
            )

        def check_if_greater_than_zero(x, v):
            self.invalid_usage(f"{x} MUST be greater than zero") if v <= 0 else None

        check_if_digit(s, start_str)
        check_if_digit(e, end_str)
        start_eps = int(start_str)
        end_eps = int(end_str)
        check_if_greater_than_zero(s, start_eps)
        check_if_greater_than_zero(e, end_eps)
        if start_eps > end_eps:
            self.invalid_usage(f"{s} CAN'T be greater than {e}")
        return start_eps, end_eps

    def invalid_usage(self, usage: str):
        print(f"Invalid Usage:\n{usage}")
        sys.exit(0)

    def validate_arg_value(self, arg: tuple[str, str], can_bes: tuple, passed: str):
        if passed not in can_bes:
            can_be = " or ".join(can_bes)
            usage = f"{arg} MUST be either {can_be}. Got '{passed}' instead"
            self.invalid_usage(usage)

    def arg_value_finder(
        self, args: list[str], target_args: tuple[str, str], default: str | None = None
    ) -> str:  # type: ignore
        for arg in target_args:
            try:
                idx = args.index(arg)
                return args[idx + 1]
            except ValueError:
                pass
        if default:
            conditional_print(f"Using default {target_args}: {default}")
            return default
        self.invalid_usage(f"Expected: {target_args}")


def test_dub_available(
    site: str, target_result: tuple[str, str] | tuple[str, str, str]
) -> tuple[bool, str]:
    test_name = "Dub availability checking"
    test_start(test_name)
    runtime_getter = get_run_time_later()
    dub_link = ""
    if site == PAHE:
        target_result = cast(tuple[str, str, str], target_result)
        dub_available = pahe.dub_available(target_result[1], target_result[2])
    else:
        dub_available, dub_link = gogo.dub_availability_and_link(target_result[0])
    rt = runtime_getter()
    if not isinstance(dub_available, bool):
        fail_test(
            test_name,
            "Boolean value",
            type(dub_available),
            rt,
            f"The returned value was: {dub_available}",
        )
    pass_test(test_name, rt)
    return dub_available, dub_link


def run_tests(args: ArgParser):
    was_hls = False
    if args.arg_in_group_was_passed(COMMANDS):
        results = test_search(args.anime_title, args.site)
        if args.verbose:
            conditional_print(f"Search Results: {results}\n")
        target_result = cast(
            tuple[str, str] | tuple[str, str, str],
            test_check_results_for_anime(args.anime_title, results),
        )
        if args.verbose:
            conditional_print(f"Target Result: {target_result}\n")
        COMMANDS.remove("search")
        if args.arg_in_group_was_passed(COMMANDS):
            dub_available, dub_link = test_dub_available(args.site, target_result)

            if args.verbose:
                conditional_print("Dub available") if dub_available else print(
                    "No Dub available"
                )
            COMMANDS.remove("dub_available")
            if args.arg_in_group_was_passed(COMMANDS):
                metadata, page_content = test_get_metadata(args.site, target_result)
                if args.verbose:
                    if args.site == GOGO:
                        conditional_print("Sub metadata")

                    def print_metadata(metadata: AnimeMetadata):
                        conditional_print(
                            f"Metadata:\nPoster Url: {metadata.poster_url}\nSummary: {metadata.summary[:100]}.. .\nEpisode Count: {metadata.episode_count}\nAiring Status: {metadata.airing_status}\nGenres: {metadata.genres}\nRelease Year: {metadata.release_year}\n"
                        )

                    print_metadata(metadata)
                COMMANDS.remove("metadata")
                if args.arg_in_group_was_passed(COMMANDS) or (
                    args.site == GOGO and args.sub_or_dub == "dub"
                ):
                    if args.end_eps > metadata.episode_count:
                        return print(
                            f"The chosen target anime has {metadata.episode_count} episodes yet you specified the ('--end_episode', '-ee') as {args.end_eps}"
                        )
                    if args.sub_or_dub == "dub" and not dub_available:
                        return print(
                            "Couldn't find Dub for the anime on the specified site"
                        )
                    elif args.site == GOGO and dub_available:
                        metadata, page_content = test_get_metadata(
                            args.site, ("", dub_link)
                        )
                        if args.verbose:
                            conditional_print("Dub metadata")
                            print_metadata(metadata)  # type: ignore
                    if args.site == PAHE:
                        target_result = cast(tuple[str, str, str], target_result)
                        episode_page_links = test_get_episode_page_links(
                            target_result[2],
                            target_result[1],
                            args.start_eps,
                            args.end_eps,
                        )
                        if args.verbose:
                            conditional_print(
                                f"Episode page links: {episode_page_links}\n"
                            )
                    else:
                        episode_page_links = []
                    COMMANDS.remove("episode_page")
                    if args.arg_in_group_was_passed(COMMANDS):
                        anime_id = (
                            gogo.extract_anime_id(page_content)
                            if args.site == GOGO
                            else 0
                        )
                        (
                            download_page_links,
                            download_info,
                        ) = test_get_download_page_links(
                            args.site,
                            args.quality,
                            args.sub_or_dub,
                            episode_page_links,
                            args.start_eps,
                            args.end_eps,
                            anime_id,
                        )
                        if args.verbose:
                            conditional_print(
                                f"Download page links: {download_page_links}\n"
                            )
                        COMMANDS.remove("download_page")
                        if args.site == PAHE and args.arg_in_group_was_passed(
                            ["download_size"]
                        ):
                            test_name = "Total Download size"
                            test_start(test_name)
                            runtime_getter = get_run_time_later()
                            total_download_size = pahe.calculate_total_download_size(
                                download_info
                            )
                            rt = runtime_getter()
                            if not isinstance(total_download_size, int):
                                fail_test(
                                    test_name,
                                    "An integer",
                                    type(total_download_size),
                                    rt,
                                    f"Bound download infos were: {download_info}",
                                )
                            pass_test(test_name, rt)
                            if args.verbose:
                                conditional_print(
                                    f"Total download size is: {total_download_size}"
                                )
                            COMMANDS.remove("download_size")
                        # HLS testing pipeine
                        if args.site == GOGO and args.arg_in_group_was_passed(
                            ["hls_links", "match_links", "segments_urls", "all"]
                        ):
                            hls_links = test_getting_hls_links(download_page_links)
                            if args.verbose:
                                conditional_print(f"HLS links: {hls_links}\n")
                            COMMANDS.remove("hls_links")
                            was_hls = True
                            if args.arg_in_group_was_passed(
                                ["segments_urls", "match_links", "all"]
                            ):
                                matched_links = test_matching_quality_to_hls_links(
                                    hls_links, args.quality
                                )
                                if args.verbose:
                                    conditional_print(
                                        f"Matched Quality links: {matched_links}\n"
                                    )
                                COMMANDS.remove("match_links")
                                if args.arg_in_group_was_passed(
                                    ["segments_urls", "all"]
                                ):
                                    segs_urls = test_getting_segments_urls(
                                        matched_links
                                    )
                                    if args.verbose:
                                        conditional_print(
                                            f"Segments URLs: {segs_urls}\n"
                                        )
                                    COMMANDS.remove("segments_urls")
                                    if args.arg_in_group_was_passed(COMMANDS):
                                        test_downloading(
                                            args.anime_title,
                                            segs_urls,
                                            True,
                                            args.start_eps,
                                            args.end_eps,
                                            args.path,
                                        )
                        if args.arg_in_group_was_passed(COMMANDS):
                            direct_download_links = test_getting_direct_download_links(
                                args.site, download_page_links, args.quality
                            )
                            if args.verbose:
                                conditional_print(f"DDLs: {direct_download_links}\n")
                            COMMANDS.remove("direct_links")
                            if args.site == GOGO and args.arg_in_group_was_passed(
                                ["download_size"]
                            ):
                                test_name = "Download size"
                                test_start(test_name)
                                runtime_getter = get_run_time_later()
                                total_download_size = gogo.CalculateTotalDowloadSize().calculate_total_download_size(
                                    direct_download_links, in_megabytes=True
                                )
                                rt = runtime_getter()
                                if not isinstance(total_download_size, int):
                                    fail_test(
                                        test_name,
                                        "An integer",
                                        type(total_download_size),
                                        rt,
                                        f"DDLs were: {direct_download_links}",
                                    )
                                pass_test(test_name, rt)
                                if args.verbose:
                                    conditional_print(
                                        f"Total download size is: {total_download_size}"
                                    )
                                COMMANDS.remove("download_size")
                            if args.arg_in_group_was_passed(COMMANDS):
                                test_downloading(
                                    args.anime_title,
                                    direct_download_links,
                                    False,
                                    args.start_eps,
                                    args.end_eps,
                                    args.path,
                                )
    if args.site == PAHE:
        return print(f"Specified {PAHE} tests passed")
    if was_hls:
        return print(f"Specified {GOGO} HLS tests passed")
    return print(f"Specified {GOGO} NORM tests passed")


if __name__ == "__main__":
    args = ArgParser(sys.argv)
    run_tests(args)




from random import choice as random_choice
from typing import Callable, cast

from bs4 import BeautifulSoup, ResultSet, Tag
from requests.cookies import RequestsCookieJar
from senpwai.utils.scraper import (
    CLIENT,
    IBYTES_TO_MBS_DIVISOR,
    PARSER,
    AnimeMetadata,
    DomainNameError,
    ProgressFunction,
    get_new_home_url_from_readme,
    match_quality,
    sanitise_title,
)
from .constants import (
    AJAX_SEARCH_URL,
    DUB_EXTENSION,
    FULL_SITE_NAME,
    AJAX_LOAD_EPS_URL,
    GOGO_HOME_URL,
    REGISTERED_ACCOUNT_EMAILS,
)

SESSION_COOKIES: RequestsCookieJar | None = None
FIRST_REQUEST = False


def search(keyword: str, ignore_dub=True) -> list[tuple[str, str]]:
    search_url = AJAX_SEARCH_URL + keyword
    response = CLIENT.get(search_url)
    content = response.json()["content"]
    soup = BeautifulSoup(content, PARSER)
    a_tags = cast(list[Tag], soup.find_all("a"))
    title_and_link: list[tuple[str, str]] = []
    for a in a_tags:
        title = a.text
        link = f'{GOGO_HOME_URL}/{a["href"]}'
        if DUB_EXTENSION in title and ignore_dub:
            continue
        title_and_link.append((title, link))
    return title_and_link


def extract_anime_id(anime_page_content: bytes) -> int:
    soup = BeautifulSoup(anime_page_content, PARSER)
    anime_id = cast(str, cast(Tag, soup.find("input", id="movie_id"))["value"])
    return int(anime_id)


def get_download_page_links(
    start_episode: int, end_episode: int, anime_id: int
) -> list[str]:
    ajax_url = AJAX_LOAD_EPS_URL.format(start_episode, end_episode, anime_id)
    content = CLIENT.get(ajax_url).content
    soup = BeautifulSoup(content, PARSER)
    a_tags = soup.find_all("a")
    download_page_links: list[str] = []
    # Reversed cause their ajax api returns the episodes in reverse order i.e 3, 2, 1
    for a in reversed(a_tags):
        resource = cast(str, a["href"]).strip()
        download_page_links.append(GOGO_HOME_URL + resource)
    return download_page_links


class GetDirectDownloadLinks(ProgressFunction):
    def __init__(self) -> None:
        super().__init__()

    def get_direct_download_links(
        self,
        download_page_links: list[str],
        user_quality: str,
        progress_update_callback: Callable = lambda _: None,
    ) -> list[str]:
        direct_download_links: list[str] = []
        for eps_pg_link in download_page_links:
            link = ""
            while not link:
                response = CLIENT.get(eps_pg_link, cookies=get_session_cookies())
                soup = BeautifulSoup(response.content, PARSER)
                a_tags = cast(
                    ResultSet[Tag],
                    cast(Tag, soup.find("div", class_="cf-download")).find_all("a"),
                )
                qualities = [a.text for a in a_tags]
                idx = match_quality(qualities, user_quality)
                redirect_link = cast(str, a_tags[idx]["href"])
                link = CLIENT.get(
                    redirect_link, cookies=get_session_cookies()
                ).headers.get("Location", redirect_link)
            direct_download_links.append(link)
            self.resume.wait()
            if self.cancelled:
                return []
            progress_update_callback(1)
        return direct_download_links


class CalculateTotalDowloadSize(ProgressFunction):
    def __init__(self):
        super().__init__()

    def calculate_total_download_size(
        self,
        direct_download_links: list[str],
        progress_update_callback: Callable = lambda _: None,
        in_megabytes=False,
    ) -> int:
        total_size = 0
        for link in direct_download_links:
            response = CLIENT.get(link, stream=True, cookies=get_session_cookies())
            size = response.headers.get("Content-Length", 0)
            if in_megabytes:
                total_size += round(int(size) / IBYTES_TO_MBS_DIVISOR)
            else:
                total_size += int(size)
            self.resume.wait()
            if self.cancelled:
                return 0
            progress_update_callback(1)
        return total_size


def get_anime_page_content(anime_page_link: str) -> bytes:
    global FIRST_REQUEST
    if FIRST_REQUEST:
        exceptions_to_ignore = [DomainNameError]
        FIRST_REQUEST = False
    else:
        exceptions_to_ignore = None
    try:
        response = CLIENT.get(
            anime_page_link, exceptions_to_ignore=exceptions_to_ignore
        )
        return response.content
    except DomainNameError:
        global GOGO_HOME_URL
        prev_home_url = GOGO_HOME_URL
        GOGO_HOME_URL = get_new_home_url_from_readme(FULL_SITE_NAME)
        return get_anime_page_content(
            anime_page_link.replace(prev_home_url, GOGO_HOME_URL)
        )


def extract_anime_metadata(anime_page_content: bytes) -> AnimeMetadata:
    soup = BeautifulSoup(anime_page_content, PARSER)
    poster_link = cast(
        str,
        cast(Tag, cast(Tag, soup.find(class_="anime_info_body_bg")).find("img"))["src"],
    )
    summary = cast(Tag, soup.find(class_="description")).get_text()
    metadata_tags = soup.find_all("p", class_="type")
    genre_tags = cast(ResultSet[Tag], metadata_tags[2].find_all("a"))
    genres = cast(list[str], [g["title"] for g in genre_tags])
    release_year = int(metadata_tags[3].get_text().replace("Released: ", ""))
    episode_count = int(
        cast(
            Tag,
            cast(
                ResultSet[Tag],
                cast(Tag, soup.find("ul", id="episode_page")).find_all("li"),
            )[-1].find("a"),
        )
        .get_text()
        .split("-")[-1]
    )
    tag = soup.find("a", title="Ongoing Anime")
    if tag:
        status = "ONGOING"
    elif episode_count == 0:
        status = "UPCOMING"
    else:
        status = "FINISHED"

    return AnimeMetadata(
        poster_link, summary, episode_count, status, genres, release_year
    )


def dub_availability_and_link(anime_title: str) -> tuple[bool, str]:
    dub_title = f"{anime_title}{DUB_EXTENSION}"
    results = search(dub_title, False)
    sanitised_dub_title = sanitise_title(dub_title, True)
    for res_title, link in results:
        if sanitised_dub_title == sanitise_title(res_title, True):
            return True, link
    return False, ""


def get_session_cookies(fresh=False) -> RequestsCookieJar:
    global SESSION_COOKIES
    if SESSION_COOKIES and not fresh:
        return SESSION_COOKIES
    login_url = f"{GOGO_HOME_URL}/login.html"
    response = CLIENT.get(login_url)
    soup = BeautifulSoup(response.content, PARSER)
    form_div = cast(Tag, soup.find("div", class_="form-login"))
    csrf_token = cast(Tag, form_div.find("input", {"name": "_csrf"}))["value"]
    form_data = {
        "email": random_choice(REGISTERED_ACCOUNT_EMAILS),
        "password": "amogus69420",
        "_csrf": csrf_token,
    }
    # A valid User-Agent is required during this post request hence the CLIENT is technically only necessary here
    response = CLIENT.post(login_url, form_data, cookies=response.cookies)
    SESSION_COOKIES = response.cookies
    if len(SESSION_COOKIES) == 0:
        return get_session_cookies()
    return SESSION_COOKIES



import os
import sys
from datetime import datetime
from pathlib import Path
from random import choice as random_choice
from subprocess import Popen
from types import TracebackType
import logging

APP_NAME = "Senpwai"
APP_NAME_LOWER = APP_NAME.lower()
VERSION = "2.1.3"
DESCRIPTION = "A desktop app for tracking and batch downloading anime"

IS_PIP_INSTALL = False
APP_EXE_PATH = ""
if getattr(sys, "frozen", False):
    # C:\Users\PC\AppData\Local\Programs\Senpwai
    ROOT_DIRECTORY = os.path.dirname(sys.executable)
    APP_EXE_PATH = sys.executable
    IS_EXECUTABLE = True
else:
    # C:\Users\PC\AppData\Local\Programs\Python\Python311\Lib\site-packages\senpwai
    ROOT_DIRECTORY = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
    # C:\Users\PC\AppData\Local\Programs\Python\Python311\Lib\site-packages
    maybe_site_packages = os.path.dirname(ROOT_DIRECTORY)
    if os.path.basename(maybe_site_packages) == "site-packages":
        IS_PIP_INSTALL = True
        # C:\Users\PC\AppData\Local\Programs\Python\Python311\Scripts\senpwai.exe
        if sys.platform == "win32":
            maybe_app_exe_path = os.path.join(
                os.path.dirname(os.path.dirname(maybe_site_packages)),
                "Scripts",
                f"{APP_NAME_LOWER}.exe",
            )
            if os.path.isfile(maybe_app_exe_path):
                APP_EXE_PATH = maybe_app_exe_path
    IS_EXECUTABLE = False
APP_EXE_ROOT_DIRECTORY = os.path.dirname(APP_EXE_PATH) if APP_EXE_PATH else ""

date = datetime.today()
IS_CHRISTMAS = True if date.month == 12 and date.day >= 20 else False

def log_exception(e: Exception):
    custom_exception_handler(type(e), e, e.__traceback__)


def custom_exception_handler(
    type_: type[BaseException], value: BaseException, traceback: TracebackType | None
):
    logging.error(f"Unhandled exception: {type_.__name__}: {value}")
    sys.__excepthook__(type_, value, traceback)


def try_deleting(path: str):
    if os.path.isfile(path):
        try:
            os.unlink(path)
        except PermissionError:
            pass


def generate_windows_setup_file_titles(app_name: str) -> tuple[str, str]:
    return (f"{app_name}-setup.exe", f"{app_name}-setup.msi")


for setup in generate_windows_setup_file_titles(APP_NAME):
    try_deleting(setup)

GITHUB_REPO_URL = "https://github.com/SenZmaKi/Senpwai"
GITHUB_API_LATEST_RELEASE_ENDPOINT = (
    "https://api.github.com/repos/SenZmaKi/Senpwai/releases/latest"
)
SEN_ANILIST_ID = 5363369
ANILIST_API_ENTRYPOINY = "https://graphql.anilist.co"
PAHE = "pahe"
GOGO = "gogo"
DUB = "dub"
SUB = "sub"
Q_1080 = "1080p"
Q_720 = "720p"
Q_480 = "480p"
Q_360 = "360p"
GOGO_NORM_MODE = "norm"
GOGO_HLS_MODE = "hls"
PAHE_NORMAL_COLOR = "#FFC300"
PAHE_HOVER_COLOR = "#FFD700"
PAHE_PRESSED_COLOR = "#FFE900"

GOGO_NORMAL_COLOR = "#00FF00"
GOGO_HOVER_COLOR = "#60FF00"
GOGO_PRESSED_COLOR = "#99FF00"

RED_NORMAL_COLOR = "#E80000"
RED_HOVER_COLOR = "#FF0202"
RED_PRESSED_COLOR = "#FF1C1C"

MINIMISED_TO_TRAY_ARG = "--minimised_to_tray"

AMOGUS_EASTER_EGG = "ඞ"


def open_folder(folder_path: str) -> None:
    if sys.platform == "win32":
        os.startfile(folder_path)
    elif sys.platform == "linux":
        Popen(["xdg-open", folder_path])
    else:
        Popen(["open", folder_path])


def requires_admin_access(folder_path):
    try:
        temp_file = os.path.join(folder_path, "temp.txt")
        open(temp_file, "w").close()
        try_deleting(temp_file)
        return False
    except PermissionError:
        return True


# Paths


assets_path = os.path.join(ROOT_DIRECTORY, "assets")


def join_from_assets(file):
    return os.path.join(assets_path, file)


misc_path = os.path.join(assets_path, "misc")


def join_from_misc(file):
    return os.path.join(misc_path, file)


SENPWAI_ICON_PATH = join_from_misc("senpwai-icon.ico")
TASK_COMPLETE_ICON_PATH = join_from_misc("task-complete.png")
LOADING_ANIMATION_PATH = join_from_misc("loading.gif")
SADGE_PIECE_PATH = join_from_misc("sadge-piece.gif")
FOLDER_ICON_PATH = join_from_misc("folder.png")

mascots_folder_path = join_from_assets("mascots")
mascots_files = list(Path(mascots_folder_path).glob("*"))
RANDOM_MACOT_ICON_PATH = (
    # Incase Senpcli was installed without Senpwai
    str(random_choice(mascots_files)) if mascots_files else ""
)


bckg_images_path = join_from_assets("background-images")


def join_from_bckg_images(
    img_title,
):  # fix windows path for Qt cause it only accepts forward slashes
    return os.path.join(bckg_images_path, img_title).replace("\\", "/")


s = "christmas.jpg" if IS_CHRISTMAS else "search.jpg"
SEARCH_WINDOW_BCKG_IMAGE_PATH = join_from_bckg_images(s)
CHOSEN_ANIME_WINDOW_BCKG_IMAGE_PATH = join_from_bckg_images("chosen-anime.jpg")
SETTINGS_WINDOW_BCKG_IMAGE_PATH = join_from_bckg_images("settings.jpg")
DOWNLOAD_WINDOW_BCKG_IMAGE_PATH = join_from_bckg_images("downloads.png")
CHOPPER_CRYING_PATH = join_from_bckg_images("chopper-crying.png")
ABOUT_BCKG_IMAGE_PATH = join_from_bckg_images("about.jpg")
UPDATE_BCKG_IMAGE_PATH = join_from_bckg_images("update.jpg")

link_icons_folder_path = join_from_assets("link-icons")


def join_from_link_icons(icon_path):
    return os.path.join(link_icons_folder_path, icon_path)


GITHUB_SPONSORS_ICON_PATH = join_from_link_icons("github-sponsors.svg")
GITHUB_ICON_PATH = join_from_link_icons("github.png")
PATREON_ICON_PATH = join_from_link_icons("patreon.png")
REDDIT_ICON_PATH = join_from_link_icons("reddit.png")
DISCORD_ICON_PATH = join_from_link_icons("discord.png")

download_icons_folder_path = join_from_assets("download-icons")


def join_from_download_icons(icon_path):
    return os.path.join(download_icons_folder_path, icon_path)


PAUSE_ICON_PATH = join_from_download_icons("pause.png")
RESUME_ICON_PATH = join_from_download_icons("resume.png")
CANCEL_ICON_PATH = join_from_download_icons("cancel.png")
REMOVE_FROM_QUEUE_ICON_PATH = join_from_download_icons("trash.png")
MOVE_UP_QUEUE_ICON_PATH = join_from_download_icons("up.png")
MOVE_DOWN_QUEUE_ICON_PATH = join_from_download_icons("down.png")

audio_folder_path = join_from_assets("audio")


def join_from_audio(audio_path):
    return os.path.join(audio_folder_path, audio_path)


GIGACHAD_AUDIO_PATH = join_from_audio("gigachad.mp3")
HENTAI_ADDICT_AUDIO_PATH = join_from_audio("aqua-crying.mp3")
MORBIUS_AUDIO_PATH = join_from_audio("morbin-time.mp3")
SEN_FAVOURITE_AUDIO_PATH = join_from_audio("sen-favourite.wav")
SEN_FAVOURITE_AUDIO_PATH = join_from_audio(
    f"one-piece-real-{random_choice((1, 2))}.mp3"
)
KAGE_BUNSHIN_AUDIO_PATH = join_from_audio("kage-bunshin-no-jutsu.mp3")
BUNSHIN_POOF_AUDIO_PATH = join_from_audio("bunshin-poof.mp3")
ZA_WARUDO_AUDIO_PATH = join_from_audio("za-warudo.mp3")
TOKI_WA_UGOKI_DASU_AUDIO_PATH = join_from_audio("toki-wa-ugoki-dasu.mp3")
WHAT_DA_HELL_AUDIO_PATH = join_from_audio("what-da-hell.mp3")
MERRY_CHRISMASU_AUDIO_PATH = join_from_audio("merry-chrismasu.mp3")

reviewer_profile_pics_folder_path = join_from_assets("reviewer-profile-pics")


def join_from_reviewer(icon_path):
    return os.path.join(reviewer_profile_pics_folder_path, icon_path)


SEN_ICON_PATH = join_from_reviewer("sen.png")
MORBIUS_IS_PEAK_ICON_PATH = join_from_reviewer("morbius-is-peak.png")
HENTAI_ADDICT_ICON_PATH = join_from_reviewer("hentai-addict.png")

navigation_bar_icons_folder_path = join_from_assets("navigation-bar-icons")


def join_from_navbar(icon_path):
    return os.path.join(navigation_bar_icons_folder_path, icon_path)


SEARCH_ICON_PATH = join_from_navbar("search.png")
DOWNLOADS_ICON_PATH = join_from_navbar("downloads.png")
SETTINGS_ICON_PATH = join_from_navbar("settings.png")
ABOUT_ICON_PATH = join_from_navbar("about.png")
UPDATE_ICON_PATH = join_from_navbar("update.png")

# Anime eastereggs

W_ANIME = {
    "vermeil",
    "golden kamuy",
    "goblin slayer",
    "hajime",
    "megalobox",
    "kengan ashura",
    "kengan asura",
    "kengan",
    "golden boy",
    "valkyrie",
    "dr stone",
    "dr. stone",
    "death parade",
    "death note",
    "code geass",
    "attack on titan",
    "kaiji",
    "shingeki no kyojin",
    "daily lives",
    "danshi koukosei",
    "daily lives of highshool boys",
    "detroit metal",
    "arakawa",
    "haikyuu",
    "kaguya",
    "chio",
    "asobi asobase",
    "prison school",
    "grand blue",
    "mob psycho",
    "to your eternity",
    "fire force",
    "mieruko",
    "fumetsu",
}
L_ANIME = {
    "tokyo ghoul",
    "sword art",
    "boku no pico",
    "full metal",
    "fmab",
    "fairy tail",
    "dragon ball",
    "hunter x hunter",
    "hunter hunter",
    "platinum end",
    "record of ragnarok",
    "7 deadly sins",
    "seven deadly sins",
    "apothecary",
}



#!/usr/bin/env python
import subprocess
import sys
from argparse import ArgumentParser, Namespace
from random import choice as random_choice
from threading import Event, Lock, Thread
from typing import Callable, cast
from queue import Queue
from os import path

from senpwai.utils.classes import SETTINGS, Anime, AnimeDetails, update_available
from senpwai.scrapers import gogo, pahe
from senpwai.utils.scraper import (
    IBYTES_TO_MBS_DIVISOR,
    Download,
    ffmpeg_is_installed,
    fuzz_str,
    lacked_episode_numbers,
    lacked_episodes,
    try_installing_ffmpeg,
)
from senpwai.utils.static import (
    IS_PIP_INSTALL,
    open_folder,
    DUB,
    APP_EXE_PATH as SENPWAI_EXE_PATH,
    GOGO,
    PAHE,
    Q_360,
    Q_480,
    Q_720,
    Q_1080,
    SUB,
    VERSION,
    GITHUB_API_LATEST_RELEASE_ENDPOINT,
    GITHUB_REPO_URL,
    ROOT_DIRECTORY,
)
from tqdm import tqdm

APP_NAME = "Senpcli"
SENPWAI_IS_INSTALLED = path.isfile(SENPWAI_EXE_PATH)
APP_NAME_LOWER = "senpcli"
DESCRIPTION = "The CLI alternative for Senpwai"
ASCII_APP_NAME = r"""
                                   .__  .__ 
  ______ ____   ____ ______   ____ |  | |__|
 /  ___// __ \ /    \\____ \_/ ___\|  | |  |
 \___ \\  ___/|   |  \  |_> >  \___|  |_|  |
/____  >\___  >___|  /   __/ \___  >____/__|
     \/     \/     \/|__|        \/         
"""

ANIME_REFERENCES = (
    "Hello friend",
    "It's called the Attack Titan",
    "Bertholdt, Reiner, Kono uragiri mono gaaaa!!!",
    "Tatakae tatake",
    "Ohio Final Boss",
    "Tokio tomare",
    "Wonder of Ohio",
    "Omoshire ore ga zangetsu da",
    "Getsuga Tenshou",
    "Rasenghan",
    "Za Warudo",
    "Star Pratina",
    "Nigurendayooo",
    "Korega jyuu da",
    "Mendokusai",
    "Dattebayo",
    "Bankaiiyeeeahhh, Ryuuumon Hooozokimaru!!!",
    "Kono asuratonkachi",
    "I devoured Barou and he devoured me right back",
    "United of States of Smaaaaash",
    "One for All Full Cowling",
    "Maid in Heaven Tokio Kasotsuru",
    "Nyaaa",
    "Pony Stark",
    "Dysfunctional Degenerate",
    "Alpha Sigma",
    "But Hey that's just a theory",
    "Bro fist",
    "King Crimson",
    "Sticky Fingers",
    "Watch Daily Lives of HighSchool Boys",
    "Watch Prison School",
    "Watch Grand Blue",
    "Watch Golden Boy, funniest shit I've ever seen",
    "Watch Isekai Ojii-san",
    "Read Kengan Asura, it's literally peak",
    "Read Dandadan",
    "Read A Thousand Splendid Suns",
    "Ryuujin no ken wo kurae!!!",
    "Ryuuga wakateki wo kurau",
    "Ookami o wagateki wo kurae",
    "Nerf this",
    "And dey say Chivalry is dead",
    "Who's next?",
    "Ryoiki Tenkai",
    "Ban.. .Kai Tensa Zangetsu",
    "Bankai Senbonzakura Kageyoshi",
    "Bankai Hihio Zabimaru",
    "Huuuero Zabimaru",
    "Nah I'd Code",
    "Senpwai: Stand proud Senpcli, you are strong",
    "We are the exception",
    "As the strongest curse Jogoat fought the fraud.. .",
    "Goodbye friend",
)


def parse_args(args: list[str]) -> Namespace:
    parser = ArgumentParser(prog=APP_NAME_LOWER, description=DESCRIPTION)
    parser.add_argument("-v", "--version", action="version", version=VERSION)
    parser.add_argument(
        "-sd",
        "--sub_or_dub",
        help="Whether to download the subbed or dubbed version of the anime",
        choices=[SUB, DUB],
        default=SETTINGS.sub_or_dub,
    )
    parser.add_argument("title", help="Title of the anime to download")
    parser.add_argument(
        "-s",
        "--site",
        help="Site to download from",
        choices=[PAHE, GOGO],
        default=SETTINGS.auto_download_site,
    )
    parser.add_argument(
        "-se",
        "--start_episode",
        type=int,
        help="Episode to start downloading at",
        default=-1,
    )
    parser.add_argument(
        "-ee",
        "--end_episode",
        type=int,
        help="Episode to stop downloading at",
        default=-1,
    )
    parser.add_argument(
        "-q",
        "--quality",
        help="Quality to download the anime in",
        choices=[Q_1080, Q_720, Q_480, Q_360],
        default=SETTINGS.quality,
    )
    parser.add_argument(
        "-hls",
        "--hls",
        help="Use HLS mode to download the anime (Gogo only and requires FFmpeg to be installed)",
        action="store_true",
    )
    parser.add_argument(
        "-sc",
        "--skip_calculating",
        help="Skip calculating the total download size (Gogo only)",
        action="store_true",
        default=SETTINGS.gogo_skip_calculate,
    )
    parser.add_argument(
        "-of",
        "--open_folder",
        help="Open the folder containing the downloaded anime once the download finishes",
        action="store_true",
    )
    parser.add_argument(
        "-msd",
        "--max_simultaneous_downloads",
        type=int,
        help="Maximum number of simultaneous downloads",
        default=SETTINGS.max_simultaneous_downloads,
    )
    return parser.parse_args(args)


def search(title: str, site: str) -> Anime | None:
    animes = (
        [
            Anime(*pahe.extract_anime_title_page_link_and_id(r))
            for r in pahe.search(title)
        ]
        if site == PAHE
        else [Anime(*r, None) for r in gogo.search(title)]
    )
    results = None
    title_fuzzy = fuzz_str(title)
    for a in animes:
        if fuzz_str(a.title) == title_fuzzy:
            results = a
            break
    if results is None:
        print(f'No anime found with the title "{title}"')
        len_animes = len(animes)
        if len_animes > 0:
            results_str = "\n".join(
                [f"{idx + 1}: {a.title}" for idx, a in enumerate(animes)]
            )
            print(
                f"Here are the results:\n{results_str}\nPick one of the results by entering its number"
            )
            num = input("> ")
            if not num.isdigit():
                print("Invalid input")
                return None
            num = int(num)
            if num < 1 or num > len_animes:
                print("Invalid input")
                return None
            results = animes[num - 1]
    return results


def validate_start_and_end_episode(
    start_episode: int, end_episode: int, total_episode_count: int
) -> tuple[int, int]:
    if end_episode == -1:
        end_episode = total_episode_count
    if end_episode > total_episode_count:
        print(
            f"Setting end episode to {total_episode_count} since the anime only has {total_episode_count} episodes"
        )
        end_episode = total_episode_count
    if start_episode > total_episode_count:
        print(
            f"Setting start episode to {total_episode_count} since the anime only has {total_episode_count} episodes"
        )
        start_episode = end_episode
    return start_episode, end_episode


def pahe_get_episode_page_links(
    start_episode: int, end_episode: int, anime_id: str, anime_page_link: str
) -> list[str]:
    (
        start_page_num,
        end_page_num,
        episode_page_count,
        first_page,
    ) = pahe.get_episode_pages_info(anime_page_link, start_episode, end_episode)
    pbar = tqdm(
        total=episode_page_count,
        desc="Getting episode page links",
        unit="eps",
        leave=False,
    )
    results = pahe.GetEpisodePageLinks().get_episode_page_links(
        start_episode,
        end_episode,
        start_page_num,
        end_page_num,
        first_page,
        anime_page_link,
        anime_id,
        pbar.update,
    )
    pbar.close()
    return results


def pahe_get_download_page_links(
    episode_page_links: list[str], quality: str, sub_or_dub: str
) -> list[str]:
    pbar = tqdm(
        total=len(episode_page_links),
        desc="Fetching download page links",
        unit="eps",
        leave=False,
    )
    (
        down_page_links,
        down_info,
    ) = pahe.GetPahewinPageLinks().get_pahewin_page_links_and_info(
        episode_page_links, pbar.update
    )
    down_page_links, down_info = pahe.bind_sub_or_dub_to_link_info(
        sub_or_dub, down_page_links, down_info
    )
    down_page_links, down_info = pahe.bind_quality_to_link_info(
        quality, down_page_links, down_info
    )
    total_download_size = pahe.calculate_total_download_size(down_info)
    pbar.close()
    print(f"Total download size: {total_download_size} MB")
    return down_page_links


def gogo_get_download_page_links(
    start_episode: int, end_episode: int, anime_page_link: str
) -> list[str]:
    anime_id = gogo.extract_anime_id(gogo.get_anime_page_content(anime_page_link))
    return gogo.get_download_page_links(start_episode, end_episode, anime_id)


def gogo_calculate_total_download_size(direct_download_links: list[str]) -> None:
    pbar = tqdm(
        total=len(direct_download_links),
        desc="Calculating total download size",
        unit="eps",
        leave=False,
    )
    total = gogo.CalculateTotalDowloadSize().calculate_total_download_size(
        direct_download_links, pbar.update
    )
    pbar.close()
    print(f"Total download size: {total // IBYTES_TO_MBS_DIVISOR} MB")


def gogo_get_direct_download_links(
    download_page_links: list[str], quality: str
) -> list[str]:
    pbar = tqdm(
        total=len(download_page_links),
        desc="Retrieving direct download links",
        unit="eps",
        leave=False,
    )
    results = gogo.GetDirectDownloadLinks().get_direct_download_links(
        download_page_links, quality, pbar.update
    )
    pbar.close()
    return results


def pahe_get_direct_download_links(download_page_links: list[str]) -> list[str]:
    pbar = tqdm(
        total=len(download_page_links),
        desc="Retrieving direct download links",
        unit="eps",
        leave=False,
    )
    results = pahe.GetDirectDownloadLinks().get_direct_download_links(
        download_page_links, pbar.update
    )
    pbar.close()
    return results


def download_thread(
    episode_title: str,
    link_or_segs_urls: str | list[str],
    anime_details: AnimeDetails,
    is_hls_download: bool,
    progress_updater: Callable,
):
    if is_hls_download:
        episode_size = len(link_or_segs_urls)
    else:
        episode_size, link_or_segs_urls = Download.get_resource_length(
            cast(str, link_or_segs_urls)
        )
    if is_hls_download:
        pbar = tqdm(
            total=episode_size,
            unit="segs",
            desc=f"Downloading [HLS] {episode_title}",
            leave=False,
        )
    else:
        pbar = tqdm(
            total=episode_size,
            unit="iB",
            unit_scale=True,
            desc=f"Downloading {episode_title}",
            leave=False,
        )
    download = Download(
        link_or_segs_urls,
        episode_title,
        anime_details.anime_folder_path,
        pbar.update,
        is_hls_download=is_hls_download,
    )
    download.start_download()
    pbar.close()
    progress_updater(1)


def download_manager(
    ddls_or_segs_urls: list[str] | list[list[str]],
    anime_details: AnimeDetails,
    is_hls_download: bool,
    max_simultaneous_downloads: int,
):
    anime_details.validate_anime_folder_path()
    desc = (
        f"Downloading [HLS] {anime_details.sanitised_title}"
        if is_hls_download
        else f"Downloading {anime_details.sanitised_title}"
    )
    episodes_pbar = tqdm(
        total=len(ddls_or_segs_urls), desc=desc, unit="eps", leave=False
    )
    download_slot_available = Event()
    download_slot_available.set()
    downloads_complete = Event()
    curr_simultaneous_downloads = 0
    update_lock = Lock()

    def update_progress(added):
        update_lock.acquire()
        nonlocal curr_simultaneous_downloads
        curr_simultaneous_downloads -= 1
        if curr_simultaneous_downloads < max_simultaneous_downloads:
            download_slot_available.set()
        episodes_pbar.update(added)
        if episodes_pbar.n == episodes_pbar.total:
            downloads_complete.set()
        update_lock.release()

    for idx, link in enumerate(ddls_or_segs_urls):
        download_slot_available.wait()
        episode_title = anime_details.episode_title(idx)
        Thread(
            target=download_thread,
            args=(episode_title, link, anime_details, is_hls_download, update_progress),
        ).start()
        curr_simultaneous_downloads += 1
        if curr_simultaneous_downloads == max_simultaneous_downloads:
            download_slot_available.clear()

    downloads_complete.wait()
    episodes_pbar.close()

    print(
        f"Download complete uWu, Senpcli ga saikou no stando da!!!\n{random_choice(ANIME_REFERENCES)}"
    )


def install_ffmpeg_prompt() -> bool:
    if not ffmpeg_is_installed():
        print(
            "HLS mode requires FFmpeg to be installed, would you like to install it? (y/n)"
        )
        if input("> ").lower() == "y":
            successfully_installed = try_installing_ffmpeg()
            if not successfully_installed:
                print("Failed to automatically install FFmpeg")
            return successfully_installed
        else:
            print("Aborting")
            return False
    return True


def already_has_all_episodes(
    anime_details: AnimeDetails,
    start_episode: int,
    end_episode: int,
    page_links: list[str],
) -> bool:
    anime_details.lacked_episode_numbers = lacked_episode_numbers(
        start_episode, end_episode, anime_details.haved_episodes
    )
    if anime_details.lacked_episode_numbers == []:
        print("Bakayorou, you already have all episodes within the provided range!!!")
        return True
    lacked_eps_page_links = lacked_episodes(
        anime_details.lacked_episode_numbers, page_links
    )
    page_links.clear()
    page_links.extend(lacked_eps_page_links)
    return False


def gogo_get_hls_links(download_page_links: list[str]) -> list[str]:
    pbar = tqdm(
        total=len(download_page_links),
        desc="Getting HLS links",
        unit="eps",
        leave=False,
    )
    results = gogo.GetHlsLinks().get_hls_links(download_page_links, pbar.update)
    pbar.close()
    return results


def gogo_get_hls_matched_quality_links(hls_links: list[str], quality: str) -> list[str]:
    pbar = tqdm(
        total=len(hls_links), desc="Matching quality to links", unit="eps", leave=False
    )
    results = gogo.GetHlsMatchedQualityLinks().get_hls_matched_quality_links(
        hls_links, quality, pbar.update
    )
    pbar.close()
    return results


def gogo_get_hls_segments_urls(matched_quality_links: list[str]) -> list[list[str]]:
    pbar = tqdm(
        total=len(matched_quality_links),
        desc="Getting segment links",
        unit="segs",
        leave=False,
    )
    results = gogo.GetHlsSegmentsUrls().get_hls_segments_urls(
        matched_quality_links, pbar.update
    )
    pbar.close()
    return results


def handle_pahe(parsed: Namespace, anime: Anime, anime_details: AnimeDetails):
    episode_page_links = pahe_get_episode_page_links(
        parsed.start_episode, parsed.end_episode, cast(str, anime.id), anime.page_link
    )
    if already_has_all_episodes(
        anime_details, parsed.start_episode, parsed.end_episode, episode_page_links
    ):
        return
    if parsed.sub_or_dub == DUB and not pahe.dub_available(
        anime.page_link, cast(str, anime.id)
    ):
        print("Dub not available for this anime")
        return
    download_page_links = pahe_get_download_page_links(
        episode_page_links, parsed.quality, parsed.sub_or_dub
    )
    direct_download_links = pahe_get_direct_download_links(download_page_links)
    download_manager(
        direct_download_links, anime_details, False, parsed.max_simultaneous_downloads
    )


def handle_gogo(parsed: Namespace, anime: Anime, anime_details: AnimeDetails):
    if parsed.hls:
        if install_ffmpeg_prompt():
            download_page_links = gogo_get_download_page_links(
                parsed.start_episode, parsed.end_episode, anime.page_link
            )
            if already_has_all_episodes(
                anime_details,
                parsed.start_episode,
                parsed.end_episode,
                download_page_links,
            ):
                return
            hls_links = gogo_get_hls_links(download_page_links)
            matched_quality_links = gogo_get_hls_matched_quality_links(
                hls_links, parsed.quality
            )
            hls_segments_urls = gogo_get_hls_segments_urls(matched_quality_links)
            download_manager(
                hls_segments_urls,
                anime_details,
                True,
                parsed.max_simultaneous_downloads,
            )
    else:
        download_page_links = gogo_get_download_page_links(
            parsed.start_episode, parsed.end_episode, anime.page_link
        )
        if already_has_all_episodes(
            anime_details, parsed.start_episode, parsed.end_episode, download_page_links
        ):
            return
        direct_download_links = gogo_get_direct_download_links(
            download_page_links, parsed.quality
        )
        if not parsed.skip_calculating:
            gogo_calculate_total_download_size(direct_download_links)
        download_manager(
            direct_download_links,
            anime_details,
            False,
            parsed.max_simultaneous_downloads,
        )


def check_for_update_thread(queue: Queue) -> None:
    is_available, download_url, file_name, _ = update_available(
        GITHUB_API_LATEST_RELEASE_ENDPOINT, APP_NAME, VERSION
    )
    queue.put((is_available, download_url, file_name))


def download_and_install_update(
    download_url: str,
    file_name: str,
) -> None:
    download_size, download_url = Download.get_resource_length(download_url)
    pbar = tqdm(
        total=download_size,
        desc="Downloading update",
        unit="iB",
        unit_scale=True,
        leave=False,
    )
    file_name_no_ext, file_ext = path.splitext(file_name)
    download = Download(
        download_url, file_name_no_ext, ROOT_DIRECTORY, pbar.update, file_ext
    )
    download.start_download()
    pbar.close()
    subprocess.Popen([path.join(ROOT_DIRECTORY, file_name), "/silent", "/update"])


def handle_update_check_result(
    is_available: bool, download_url, file_name: str
) -> None:
    if not is_available:
        return
    if IS_PIP_INSTALL:
        return print(
            'Update available, install it by running "pip install senpwai --upgrade"'
        )
    if SENPWAI_IS_INSTALLED:
        return print("Update available, install it by updating Senpwai")
    if sys.platform == "win32":
        print("Update available, would you like to download and install it? (y/n)")
        if input("> ").lower() == "y":
            try:
                download_and_install_update(download_url, file_name)
            except PermissionError:
                print(
                    "Retry in an admin shell e.g., if you're using Command Prompt run it as an administrator"
                )
        return
    print(
        f"A new version is available, but to install it you'll have to build from source\nThere is a guide at: {GITHUB_REPO_URL}"
    )


def start_update_check_thread() -> tuple[Thread, Queue]:
    update_check_result_queue = Queue()
    update_check_thread = Thread(
        target=check_for_update_thread, args=(update_check_result_queue,)
    )
    update_check_thread.start()
    return update_check_thread, update_check_result_queue


def get_anime_and_anime_details(parsed) -> tuple[Anime, AnimeDetails] | None:
    anime = search(parsed.title, parsed.site)
    if anime is None:
        return None
    if parsed.sub_or_dub == DUB and parsed.site == GOGO:
        dub_available, anime.page_link = gogo.dub_availability_and_link(anime.title)
        if not dub_available:
            print("Dub not available for this anime")
            return None
    anime_details = AnimeDetails(anime, parsed.site)
    if parsed.start_episode == -1:
        if (
            anime_details.haved_end is not None
            and anime_details.haved_end < anime_details.metadata.episode_count
        ):
            parsed.start_episode = anime_details.haved_end + 1
        else:
            parsed.start_episode = 1
    parsed.start_episode, parsed.end_episode = validate_start_and_end_episode(
        parsed.start_episode, parsed.end_episode, anime_details.metadata.episode_count
    )
    return anime, anime_details


def initiate_download_pipeline(
    parsed: Namespace, anime: Anime, anime_details: AnimeDetails
):
    print(f"Downloading to: {anime_details.anime_folder_path}")
    if parsed.site == PAHE:
        handle_pahe(parsed, anime, anime_details)
    else:
        handle_gogo(parsed, anime, anime_details)
    if parsed.open_folder:
        open_folder(anime_details.anime_folder_path)


def validate_args(parsed: Namespace) -> bool:
    if parsed.end_episode < parsed.start_episode:
        print("End episode cannot be less than start episode, hontoni baka ga")
        return False
    elif parsed.start_episode < 1 and parsed.start_episode != -1:
        print("Start episode cannot be less than 1, is that your IQ?")
        return False
    elif parsed.end_episode < 1 and parsed.end_episode != -1:
        print("End episode cannot be less than 1, is that your brain cell count?")
        return False
    elif parsed.site != GOGO and parsed.hls:
        print("Setting site to Gogo since HLS mode is only available for Gogo")
        parsed.site = GOGO

    return True


def main():
    try:
        args = sys.argv[1:]
        print(ASCII_APP_NAME)
        parsed = parse_args(args)
        if not validate_args(parsed):
            return
        update_check_thread, update_check_result_queue = start_update_check_thread()
        anime_and_anime_details = get_anime_and_anime_details(parsed)
        if anime_and_anime_details is None:
            return
        initiate_download_pipeline(parsed, *anime_and_anime_details)
        update_check_thread.join()
        handle_update_check_result(*update_check_result_queue.get())

    except KeyboardInterrupt:
        print("\n\nAborted")


if __name__ == "__main__":
    main()



import re
import math
from typing import Any, Callable, cast
from requests import Response
from bs4 import BeautifulSoup, Tag
from senpwai.utils.scraper import (
    CLIENT,
    PARSER,
    AnimeMetadata,
    DomainNameError,
    ProgressFunction,
    get_new_home_url_from_readme,
    match_quality,
)
from .constants import (
    PAHE_HOME_URL,
    FULL_SITE_NAME,
    API_ENTRY_POINT,
    ANIME_PAGE_URL,
    LOAD_EPISODES_URL,
    DUB_PATTERN,
    EPISODE_PAGE_URL,
    EPISODE_SIZE_REGEX,
    KWIK_PAGE_REGEX,
    PARAM_REGEX,
)

FIRST_REQUEST = True
COOKIES = {"__ddg1_": "", "__ddg2_": ""}
"""
For some reason these cookies just need to be set as in they don't even need to be valid
If something crashes, try updating to something like: 
COOKIES = {
    "__ddg1_": f"; Expires=Tue, 19 Jan 2038 03:14:07 GMT; Domain={PAHE_DOMAIN}; Path=/",
    "__ddg2_": f"; Expires=Tue, 19 Jan 2038 03:14:07 GMT; Domain={PAHE_DOMAIN}; Path=/",
}
"""

def site_request(url: str) -> Response:
    """
    For requests that go specifically to the domain animepahe.ru instead of e.g., pahe.win or kwik.si
    Typically these requests need the cookies
    """
    try:
        # We only want to handle the domain change incase this is the first request
        # This is to avoid raising DomainNameError when the something else broke instead
        global FIRST_REQUEST
        if FIRST_REQUEST:
            exceptions_to_ignore = [DomainNameError]
            FIRST_REQUEST = False
        else:
            exceptions_to_ignore = None
        response = CLIENT.get(
            url, cookies=COOKIES, exceptions_to_ignore=exceptions_to_ignore
        )
        COOKIES.update(response.cookies)
    except DomainNameError:
        global PAHE_HOME_URL
        PAHE_HOME_URL = get_new_home_url_from_readme(FULL_SITE_NAME)
        return site_request(url)
    return response


def search(keyword: str) -> list[dict[str, str]]:
    search_url = f"{API_ENTRY_POINT}search&q={keyword}"
    response = site_request(search_url)
    decoded = cast(dict, response.json())
    # The search api endpoint won't return json containing the data key if no results are found
    return decoded.get("data", [])


def extract_anime_title_page_link_and_id(
    result: dict[str, str],
) -> tuple[str, str, str]:
    anime_id = result["session"]
    title = result["title"]
    page_link = ANIME_PAGE_URL.format(anime_id)
    return title, page_link, anime_id


def get_episode_pages_info(
    anime_page_link: str, start_episode: int, end_episode: int
) -> tuple[int, int, int, dict[str, Any]]:
    page_url = LOAD_EPISODES_URL.format(anime_page_link, 1)
    decoded = site_request(page_url).json()
    per_page: int = decoded["per_page"]
    start_page = math.ceil(start_episode / per_page)
    end_page = math.ceil(end_episode / per_page)
    episode_page_count = (end_page - start_page) + 1
    return start_page, end_page, episode_page_count, decoded


class GetEpisodePageLinks(ProgressFunction):
    def __init__(self) -> None:
        super().__init__()

    @staticmethod
    def generate_episode_page_links(
        start_episode: int,
        end_episode: int,
        episodes_data: list[dict[str, Any]],
        anime_id: str,
    ):
        # These values should theoritically never remain as they are unless something crashes
        # as in both of the ifs in the for loop should always eventually evaluate to True
        start_idx = 0
        end_idx = 0

        for idx, episode in enumerate(episodes_data):
            if episode["episode"] == start_episode:
                start_idx = idx
            if episode["episode"] == end_episode:
                end_idx = idx
                break
        episodes_data = episodes_data[start_idx : end_idx + 1]
        episode_sessions = [episode["session"] for episode in episodes_data]
        return [
            EPISODE_PAGE_URL.format(anime_id, episode_session)
            for episode_session in episode_sessions
        ]

    # Retrieves a list of the episode page links (not download links)
    def get_episode_page_links(
        self,
        start_episode: int,
        end_episode: int,
        start_page_num: int,
        end_page_num: int,
        first_page: dict[str, Any],
        anime_page_link: str,
        anime_id: str,
        progress_update_callback: Callable = lambda _: None,
    ) -> list[str]:
        page_url = anime_page_link
        episodes_data: list[dict[str, Any]] = []
        if start_page_num == 1:
            episodes_data.extend(first_page["data"])
            start_page_num += 1
            progress_update_callback(1)
        for page_num in range(start_page_num, end_page_num + 1):
            page_url = LOAD_EPISODES_URL.format(anime_page_link, page_num)
            decoded = site_request(page_url).json()
            # To avoid episodes like 7.5 and 5.5 cause they're usually just recaps
            episodes = [ep for ep in decoded["data"] if isinstance(ep["episode"], int)]
            episodes_data.extend(episodes)
            page_url = decoded["next_page_url"]
            self.resume.wait()
            if self.cancelled:
                return []
            progress_update_callback(1)
        return GetEpisodePageLinks.generate_episode_page_links(
            start_episode, end_episode, episodes_data, anime_id
        )


class GetPahewinPageLinks(ProgressFunction):
    def __init__(self) -> None:
        super().__init__()

    def get_pahewin_page_links_and_info(
        self,
        episode_page_links: list[str],
        progress_update_callback: Callable = lambda _: None,
    ) -> tuple[list[list[str]], list[list[str]]]:
        pahewin_links: list[list[str]] = []
        download_info: list[list[str]] = []
        for episode_page_link in episode_page_links:
            page_content = site_request(episode_page_link).content
            soup = BeautifulSoup(page_content, PARSER)
            pahewin_data = soup.find_all("a", class_="dropdown-item", target="_blank")
            if pahewin_data:
                pahewin_links.append([cast(str, link["href"]) for link in pahewin_data])
                download_info.append([li.text.strip() for li in pahewin_data])
            self.resume.wait()
            if self.cancelled:
                return ([], [])
            progress_update_callback(1)
        return (pahewin_links, download_info)


def is_dub(anime_info: str) -> bool:
    return anime_info.endswith(DUB_PATTERN)


def dub_available(anime_page_link: str, anime_id: str) -> bool:
    page_url = LOAD_EPISODES_URL.format(anime_page_link, 1)
    decoded = site_request(page_url).json()
    episodes_data = decoded.get("data", None)
    if episodes_data is None:
        return False
    episode_sessions = [episode["session"] for episode in episodes_data]
    episode_links = [
        EPISODE_PAGE_URL.format(anime_id, episode_session)
        for episode_session in episode_sessions
    ]
    (
        _,
        download_info,
    ) = GetPahewinPageLinks().get_pahewin_page_links_and_info(episode_links[:1])

    for info in download_info[0]:
        if is_dub(info):
            return True
    return False


def bind_sub_or_dub_to_link_info(
    sub_or_dub: str,
    pahewin_download_page_links: list[list[str]],
    download_info: list[list[str]],
) -> tuple[list[list[str]], list[list[str]]]:
    bound_links: list[list[str]] = []
    bound_info: list[list[str]] = []
    for link_list, episode_info in zip(pahewin_download_page_links, download_info):
        links: list[str] = []
        infos: list[str] = []
        for link, info in zip(link_list, episode_info):
            is_dub_link = is_dub(info)
            if (sub_or_dub == "sub" and not is_dub_link) or (
                sub_or_dub == "dub" and is_dub_link
            ):
                links.append(link)
                infos.append(info)
        if links and infos:
            bound_links.append(links)
            bound_info.append(infos)
    return (bound_links, bound_info)


def bind_quality_to_link_info(
    quality: str,
    pahewin_download_page_links: list[list[str]],
    download_info: list[list[str]],
) -> tuple[list[str], list[str]]:
    bound_links: list[str] = []
    bound_info: list[str] = []
    for links, infos in zip(pahewin_download_page_links, download_info):
        index = match_quality(infos, quality)
        bound_links.append(links[index])
        bound_info.append(infos[index])
    return (bound_links, bound_info)


def calculate_total_download_size(bound_info: list[str]) -> int:
    total_size = 0
    download_sizes: list[int] = []
    for episode in bound_info:
        match = cast(re.Match, EPISODE_SIZE_REGEX.search(episode))
        size = int(match.group(1))
        download_sizes.append(size)
        total_size += size
    return total_size


def get_string(content: str, s1: int) -> int:
    map_thing = "0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ+/"
    s2 = 10
    map_string = map_thing[:s2]
    acc = 0
    for index, c in enumerate(reversed(content)):
        acc += (int(c) if c.isdigit() else 0) * int(math.pow(s1, index))
    k = ""
    while acc > 0:
        k = map_string[acc % s2] + k
        acc = (acc - (acc % s2)) // s2
    return int(k) if k.isdigit() else 0


# Courtesy of Saikou app https://github.com/saikou-app/saikou
def decrypt_token_and_post_url_page(full_key: str, key: str, v1: int, v2: int) -> str:
    r = ""
    i = 0
    while i < len(full_key):
        s = ""
        while full_key[i] != key[v2]:
            s += full_key[i]
            i += 1
        for j in range(len(key)):
            s = s.replace(key[j], str(j))
        r += chr(get_string(s, v2) - v1)
        i += 1
    return r


class GetDirectDownloadLinks(ProgressFunction):
    def __init__(self) -> None:
        super().__init__()

    def get_direct_download_links(
        self,
        pahewin_download_page_links: list[str],
        progress_update_callback: Callable = lambda _: None,
    ) -> list[str]:
        direct_download_links: list[str] = []
        for pahewin_link in pahewin_download_page_links:
            # Extract kwik page links
            html_page = CLIENT.get(pahewin_link).text
            download_link = cast(
                re.Match[str], KWIK_PAGE_REGEX.search(html_page)
            ).group()

            # Extract direct download links from kwik page links
            response = CLIENT.get(download_link)
            cookies = response.cookies
            match = cast(re.Match, PARAM_REGEX.search(response.text))
            full_key, key, v1, v2 = (
                match.group(1),
                match.group(2),
                match.group(3),
                match.group(4),
            )
            decrypted = decrypt_token_and_post_url_page(full_key, key, int(v1), int(v2))
            soup = BeautifulSoup(decrypted, PARSER)
            post_url = cast(str, cast(Tag, soup.form)["action"])
            token_value = cast(str, cast(Tag, soup.input)["value"])
            response = CLIENT.post(
                post_url,
                headers=CLIENT.append_headers({"Referer": download_link}),
                cookies=cookies,
                data={"_token": token_value},
                allow_redirects=False,
            )
            direct_download_link = response.headers["Location"]
            direct_download_links.append(direct_download_link)
            self.resume.wait()
            if self.cancelled:
                return []
            progress_update_callback(1)
        return direct_download_links


def get_anime_metadata(anime_id: str) -> AnimeMetadata:
    page_link = f"{PAHE_HOME_URL}/anime/{anime_id}"
    page_content = site_request(page_link).content
    soup = BeautifulSoup(page_content, PARSER)
    poster = soup.find(class_="youtube-preview")
    if not isinstance(poster, Tag):
        poster = cast(Tag, soup.find(class_="poster-image"))
    poster_link = cast(str, poster["href"])
    summary = cast(Tag, soup.find(class_="anime-synopsis")).get_text()
    genres_tags = cast(
        Tag, cast(Tag, soup.find(class_="anime-genre font-weight-bold")).find("ul")
    ).find_all("li")
    genres: list[str] = []
    for genre in genres_tags:
        genres.append(cast(str, cast(Tag, genre.find("a"))["title"]))
    season_and_year = cast(
        str, cast(Tag, soup.select_one('a[href*="/anime/season/"]'))["title"]
    )
    _, release_year = season_and_year.split(" ")
    page_link = ANIME_PAGE_URL.format(anime_id)
    decoded = site_request(page_link).json()
    episode_count = decoded["total"]
    tag = soup.find(title="Currently Airing")
    if tag:
        status = "ONGOING"
    elif episode_count == 0:
        status = "UPCOMING"
    else:
        status = "FINISHED"
    return AnimeMetadata(
        poster_link, summary, episode_count, status, genres, int(release_year)
    )



import re

GOGO = "gogo"
GOGO_HOME_URL = "https://anitaku.to"
AJAX_SEARCH_URL = "https://ajax.gogo-load.com/site/loadAjaxSearch?keyword="
AJAX_LOAD_EPS_URL = (
    "https://ajax.gogo-load.com/ajax/load-list-episode?ep_start={}&ep_end={}&id={}"
)
FULL_SITE_NAME = "Gogoanime"
DUB_EXTENSION = " (Dub)"
REGISTERED_ACCOUNT_EMAILS = [
    "benida7218@weirby.com",
    "hareki4411@wisnick.com",
    "nanab67795@weirby.com",
    "xener53725@weirby.com",
    "nenado3105@weirby.com",
    "yaridod257@weirby.com",
    "ketoh33964@weirby.com",
    "kajade1254@wisnick.com",
    "nakofe3005@weirby.com",
    "gedidij506@weirby.com",
    "sihaci1525@undewp.com",
    "lorimob952@soebing.com",
    "nigeha6048@undewp.com",
    "goriwij739@undewp.com",
    "pekivik280@soebing.com",
    "hapas66158@undewp.com",
    "wepajof522@undewp.com",
    "semigo1458@undewp.com",
    "xojecog864@undewp.com",
    "lobik97135@wanbeiz.com",
]

KEYS_REGEX = re.compile(rb"(?:container|videocontent)-(\d+)")
ENCRYPTED_DATA_REGEX = re.compile(rb'data-value="(.+?)"')



from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from cryptography.hazmat.primitives.padding import PKCS7
from .constants import KEYS_REGEX, ENCRYPTED_DATA_REGEX
from senpwai.utils.scraper import (
    PARSER,
    CLIENT,
    ProgressFunction,
    match_quality,
)
from bs4 import BeautifulSoup, Tag
from typing import Any, cast, Callable
from yarl import URL
import base64
import re
import json


class GetHlsMatchedQualityLinks(ProgressFunction):
    def __init__(self) -> None:
        super().__init__()

    def get_hls_matched_quality_links(
        self,
        hls_links: list[str],
        quality: str,
        progress_update_callback: Callable = lambda _: None,
    ) -> list[str]:
        matched_links: list[str] = []
        for link in hls_links:
            response = CLIENT.get(link)
            self.resume.wait()
            if self.cancelled:
                return []
            lines = response.text.split(",")
            qualities = [line for line in lines if "NAME=" in line]
            idx = match_quality(qualities, quality)
            resource = qualities[idx].splitlines()[1]
            base_url = URL(link).parent
            matched_links.append(f"{base_url}/{resource}")
            progress_update_callback(1)
        return matched_links


class GetHlsSegmentsUrls(ProgressFunction):
    def __init__(self) -> None:
        super().__init__()

    def get_hls_segments_urls(
        self,
        matched_links: list[str],
        progress_update_callback: Callable = lambda _: None,
    ) -> list[list[str]]:
        segments_urls: list[list[str]] = []
        for link in matched_links:
            response = CLIENT.get(link)
            self.resume.wait()
            if self.cancelled:
                return []
            segments = response.text.splitlines()
            base_url = "/".join(link.split("/")[:-1])
            segment_urls: list[str] = []
            for seg in segments:
                if seg.endswith(".ts"):
                    segment_url = f"{base_url}/{seg}"
                    segment_urls.append(segment_url)
            segments_urls.append(segment_urls)
            progress_update_callback(1)
        return segments_urls


class GetHlsLinks(ProgressFunction):
    def __init__(self) -> None:
        super().__init__()

    def get_hls_links(
        self,
        download_page_links: list[str],
        progress_update_callback: Callable[[int], Any] = lambda _: None,
    ) -> list[str]:
        hls_links: list[str] = []
        for eps_url in download_page_links:
            hls_links.append(extract_stream_url(get_embed_url(eps_url)))
            self.resume.wait()
            if self.cancelled:
                return []
            progress_update_callback(1)
        return hls_links


def get_embed_url(episode_page_link: str) -> str:
    response = CLIENT.get(episode_page_link)
    soup = BeautifulSoup(response.content, PARSER)
    return cast(str, cast(Tag, soup.select_one("iframe"))["src"])


def aes_encrypt(data: str, *, key, iv) -> bytes:
    backend = default_backend()
    cipher = Cipher(algorithms.AES(key), modes.CBC(iv), backend=backend)
    encryptor = cipher.encryptor()
    padder = PKCS7(128).padder()
    padded_data = padder.update(data.encode()) + padder.finalize()
    return base64.b64encode(encryptor.update(padded_data) + encryptor.finalize())


def aes_decrypt(data: str, *, key, iv) -> str:
    backend = default_backend()
    cipher = Cipher(algorithms.AES(key), modes.CBC(iv), backend=backend)
    decryptor = cipher.decryptor()
    decrypted_data = decryptor.update(base64.b64decode(data)) + decryptor.finalize()
    unpadder = PKCS7(128).unpadder()
    unpadded_data = unpadder.update(decrypted_data) + unpadder.finalize()
    return unpadded_data.decode()


def extract_stream_url(embed_url: str) -> str:
    parsed_url = URL(embed_url)
    content_id = parsed_url.query["id"]
    streaming_page_host = f"https://{parsed_url.host}/"
    streaming_page = CLIENT.get(embed_url).content

    encryption_key, iv, decryption_key = (
        _.group(1) for _ in KEYS_REGEX.finditer(streaming_page)
    )
    component = aes_decrypt(
        cast(re.Match[bytes], ENCRYPTED_DATA_REGEX.search(streaming_page))
        .group(1)
        .decode(),
        key=encryption_key,
        iv=iv,
    ) + "&id={}&alias={}".format(
        aes_encrypt(content_id, key=encryption_key, iv=iv).decode(), content_id
    )

    component = component.split("&", 1)[1]
    ajax_response = CLIENT.get(
        streaming_page_host + "encrypt-ajax.php?" + component,
        headers=CLIENT.append_headers({"X-Requested-With": "XMLHttpRequest"}),
    )
    content = json.loads(
        aes_decrypt(ajax_response.json()["data"], key=decryption_key, iv=iv)
    )

    try:
        source = content["source"]
        stream_url = source[0]["file"]
    except KeyError:
        source = content["source_bk"]
        stream_url = source[0]["file"]
    return stream_url



import sys
from time import time
from typing import Callable, cast

from PyQt6.QtCore import QEvent, QMutex, QObject, QSize, Qt, QTimer, QUrl
from PyQt6.QtGui import (
    QIcon,
    QKeyEvent,
    QMovie,
    QPaintEvent,
    QPainter,
    QPainterPath,
    QPen,
    QPixmap,
)
from PyQt6.QtMultimedia import QAudioOutput, QMediaPlayer
from PyQt6.QtWidgets import (
    QFrame,
    QHBoxLayout,
    QLabel,
    QLineEdit,
    QProgressBar,
    QPushButton,
    QScrollArea,
    QSizePolicy,
    QTextBrowser,
    QVBoxLayout,
    QWidget,
)

from senpwai.utils.classes import Settings
from senpwai.utils.static import (
    FOLDER_ICON_PATH,
    GOGO_NORM_MODE,
    GOGO_NORMAL_COLOR,
    PAHE_NORMAL_COLOR,
    PAHE_PRESSED_COLOR,
    RED_NORMAL_COLOR,
    RED_PRESSED_COLOR,
    open_folder,
)


def set_minimum_size_policy(object):
    object.setSizePolicy(QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Minimum)
    object.setFixedSize(object.sizeHint())


def fix_qt_path_for_windows(path: str) -> str:
    if sys.platform == "win32":
        path = path.replace("/", "\\")
    return path


class BckgImg(QLabel):
    def __init__(self, parent: QWidget, image_path: str) -> None:
        super().__init__(parent)
        pixmap = QPixmap(image_path)
        self.setPixmap(pixmap)
        self.setScaledContents(True)


class Animation(QLabel):
    def __init__(
        self,
        animation_path: str,
        animation_size_x: int,
        animation_size_y: int,
        parent: QWidget | None = None,
    ):
        super().__init__(parent)
        self.animation = QMovie(animation_path)
        self.animation.setScaledSize(QSize(animation_size_x, animation_size_y))
        self.setMovie(self.animation)

    def start(self):
        return self.animation.start()

    def stop(self):
        return self.animation.stop()


class StyledLabel(QLabel):
    def __init__(
        self,
        parent=None,
        font_size: int = 20,
        bckg_color: str = "rgba(0, 0, 0, 220)",
        border_radius=10,
        font_color="white",
    ):
        super().__init__(parent)
        self.setStyleSheet(
            f"""
                    QLabel {{
                        color: {font_color};
                        font-size: {font_size}px;
                        font-family: "Berlin Sans FB Demi";
                        background-color: {bckg_color};
                        border-radius: {border_radius}px;
                        border: 1px solid black;
                        padding: 5px;
                    }}
                            """
        )


class StyledTextBrowser(QTextBrowser):
    def __init__(
        self,
        parent=None,
        font_size: int = 20,
        bckg_color: str = "rgba(0, 0, 0, 220)",
        border_radius=10,
        font_color="white",
    ):
        super().__init__(parent)
        self.setOpenExternalLinks(True)
        self.setStyleSheet(
            f"""
                    QTextEdit {{
                        color: {font_color};
                        font-size: {font_size}px;
                        font-family: "Berlin Sans FB Demi";
                        background-color: {bckg_color};
                        border-radius: {border_radius}px;
                        border: 1px solid black;
                        padding: 5px;
                    }}
                            """
        )


class StyledButton(QPushButton):
    def __init__(
        self,
        parent: QWidget | None,
        font_size: int,
        font_color: str,
        normal_color: str,
        hover_color: str,
        pressed_color: str,
        border_radius=12,
    ):
        super().__init__()
        if parent:
            self.setParent(parent)
        self.setCursor(Qt.CursorShape.PointingHandCursor)
        self.setStyleSheet(
            f"""
            QPushButton {{
                color: {font_color};
                background-color: {normal_color};
                border: 1px solid black;
                font-size: {font_size}px;
                font-family: "Berlin Sans FB Demi";
                padding: 10px;
                border-radius: {border_radius}px;
            }}
            QPushButton:hover {{
                background-color: {hover_color};
            }}
            QPushButton:pressed {{  
                background-color: {pressed_color};
            }}
        """
        )
        self.installEventFilter(self)

    def eventFilter(self, a0: QObject | None, a1: QEvent | None):
        if a0 == self:
            if (
                isinstance(a1, QKeyEvent)
                and a1.type() == a1.Type.KeyPress
                and (a1.key() in (Qt.Key.Key_Return, Qt.Key.Key_Enter))
            ):
                self.animateClick()
                return True

        return super().eventFilter(a0, a1)


class DualStateButton(StyledButton):
    def __init__(
        self,
        parent: QWidget | None,
        font_size: int,
        font_color: str,
        hover_color: str,
        set_color: str,
        on_text: str,
        off_text: str,
        unset_color="rgba(128, 128, 128, 255)",
        border_radius=12,
    ):
        super().__init__(
            parent,
            font_size,
            font_color,
            unset_color,
            hover_color,
            set_color,
            border_radius,
        )
        self.on = False
        self.on_text = on_text
        self.off_text = off_text
        self.setText(on_text)
        self.off_stylesheet = self.styleSheet()
        styles_to_overwride = f"""
            QPushButton {{
            border-radius: 7px;
            background-color: {set_color};
        }}
        QPushButton:hover {{
            background-color: {hover_color};
        }}
    """
        self.on_stylesheet = self.off_stylesheet + styles_to_overwride
        self.clicked.connect(self.change_status)

    def change_status(self):
        self.on = not self.on
        if self.on:
            self.setStyleSheet(self.on_stylesheet)
            self.setText(self.off_text)
        else:
            self.setStyleSheet(self.off_stylesheet)
            self.setText(self.on_text)
        set_minimum_size_policy(self)


class OptionButton(StyledButton):
    def __init__(
        self,
        parent: QWidget | None,
        option: Settings.types,
        option_displayed: str,
        font_size: int,
        chosen_color: str,
        hover_color: str,
        font_color="white",
    ):
        super().__init__(
            parent,
            font_size,
            font_color,
            "rgba(128, 128, 128, 255)",
            chosen_color,
            hover_color,
        )
        self.not_picked_style_sheet = self.styleSheet()
        styles_to_overwride = f"""
            QPushButton {{
            border-radius: 7px;
            background-color: {chosen_color};
        }}
        QPushButton:hover {{
            background-color: {chosen_color};
        }}
    """
        self.picked_style_sheet = self.not_picked_style_sheet + styles_to_overwride
        self.option = option
        self.setText(option_displayed)
        self.clicked.connect(lambda: self.set_picked_status(True))

    def set_picked_status(self, picked: bool):
        if picked:
            self.setStyleSheet(self.picked_style_sheet)
        else:
            self.setStyleSheet(self.not_picked_style_sheet)


class Icon(QIcon):
    def __init__(self, x: int, y: int, icon_path: str):
        self.x = x
        self.y = y
        pixmap = QPixmap(icon_path).scaled(
            self.x, self.y, Qt.AspectRatioMode.IgnoreAspectRatio
        )
        super().__init__(pixmap)


class IconButton(QPushButton):
    def __init__(
        self, icon: Icon, size_factor: int | float, parent: QWidget | None = None
    ):
        super().__init__(parent)
        self.setFixedSize(icon.x, icon.y)
        self.setIcon(icon)
        self.setIconSize(self.size() / size_factor)
        self.enterEvent = lambda event: self.setIconSize(QSize(icon.x, icon.y))
        self.leaveEvent = lambda a0: self.setIconSize(
            QSize(icon.x, icon.y) / size_factor
        )
        self.setCursor(Qt.CursorShape.PointingHandCursor)
        self.setStyleSheet(
            """
            QPushButton {
                border: none; 
                background-color: transparent;
            }"""
        )
        self.installEventFilter(self)

    def eventFilter(self, a0: QObject | None, a1: QEvent | None):
        if a0 == self:
            if (
                isinstance(a1, QKeyEvent)
                and a1.type() == a1.Type.KeyPress
                and ((a1.key() == Qt.Key.Key_Return) or (a1.key() == Qt.Key.Key_Enter))
            ):
                if isinstance(a0, QPushButton):
                    self.animateClick()
                    return True

        return super().eventFilter(a0, a1)


class AnimationAndText(QWidget):
    def __init__(
        self,
        animation_path: str,
        animation_size_x: int,
        animation_size_y: int,
        text: str,
        paint_x: int,
        paint_y: int,
        font_size: int,
        parent: QWidget | None = None,
    ):
        super().__init__(parent)
        self.animation = Animation(
            animation_path, animation_size_x, animation_size_y, parent
        )
        self.animation_path = animation_path
        self.text_label = OutlinedLabel(parent, paint_x, paint_y)
        self.text_label.setText(text)
        self.text_label.setStyleSheet(
            f"""
                    OutlinedLabel {{
                        color: #FFEF00;
                        font-size: {font_size}px;
                        font-family: "Berlin Sans FB Demi";
                        }}
                        """
        )
        self.main_layout = QVBoxLayout()
        self.main_layout.addWidget(self.animation)
        self.main_layout.addWidget(self.text_label)
        self.main_layout.setAlignment(self.animation, Qt.AlignmentFlag.AlignCenter)
        self.main_layout.setAlignment(self.text_label, Qt.AlignmentFlag.AlignCenter)
        self.setLayout(self.main_layout)

    def start(self):
        return self.animation.start()

    def stop(self):
        return self.animation.stop()


class OutlinedLabel(QLabel):
    def __init__(self, parent: QWidget | None, paint_x: int, paint_y: int):
        self.paint_x = paint_x
        self.paint_y = paint_y
        super().__init__(parent)

    def paintEvent(self, a0: QPaintEvent | None):
        painter = QPainter(self)
        painter.setRenderHint(QPainter.RenderHint.Antialiasing)

        # Draw the outline around the text
        pen = QPen()
        pen.setWidth(5)
        painter.setPen(pen)

        path = QPainterPath()
        path.addText(self.paint_x, self.paint_y, self.font(), self.text())
        painter.drawPath(path)

        # Call the parent class's paintEvent to draw the label background and other properties
        painter.end()
        return super().paintEvent(a0)


class OutlinedButton(StyledButton):
    def __init__(
        self,
        paint_x,
        paint_y,
        parent: QWidget | None,
        font_size: int,
        font_color: str,
        normal_color: str,
        hover_color: str,
        pressed_color: str,
        border_radius=12,
    ):
        self.paint_x = paint_x
        self.paint_y = paint_y
        super().__init__(
            parent,
            font_size,
            font_color,
            normal_color,
            hover_color,
            pressed_color,
            border_radius,
        )

    def paintEvent(self, a0: QPaintEvent | None):
        painter = QPainter(self)
        painter.setRenderHint(QPainter.RenderHint.Antialiasing)

        # Draw the outline around the text
        pen = QPen()
        pen.setWidth(5)
        painter.setPen(pen)

        path = QPainterPath()
        path.addText(self.paint_x, self.paint_y, self.font(), self.text())
        painter.drawPath(path)

        # Call the parent class's paintEvent to draw the button background and other properties
        painter.end()
        return super().paintEvent(a0)


class AudioPlayer(QMediaPlayer):
    def __init__(self, parent: QWidget | None, audio_path: str, volume: int = 50):
        super().__init__(parent)
        audio_output = QAudioOutput(parent)
        self.setAudioOutput(audio_output)
        self.setSource(QUrl.fromLocalFile(audio_path))
        audio_output.setVolume(volume)


class ErrorLabel(StyledLabel):
    def __init__(
        self,
        font_size: int,
        shown_duration_in_secs: int = 3,
        parent: QWidget | None = None,
    ):
        super().__init__(parent, font_size, font_color="red")
        self.shown_duration_in_secs = shown_duration_in_secs
        self.timer = QTimer(self)

    def show(self):
        if not self.timer.isActive():
            self.timer.setSingleShot(True)
            self.timer.timeout.connect(self.hide)
            self.timer.start(self.shown_duration_in_secs * 1000)
            return super().show()


class ProgressBarWithoutButtons(QWidget):
    ongoing_stylesheet = f"""
            QProgressBar {{
                border: 1px solid black;
                color: black;
                text-align: center;
                border-radius: 10px;
                background-color: rgba(255, 255, 255, 150);
                font-size: 22px;
                font-family: "Berlin Sans FB Demi";
            }}

            QProgressBar::chunk {{
                background-color: {PAHE_NORMAL_COLOR};
                border-radius: 10px;
            }}
        """
    styles_to_overwride = """
            QProgressBar::chunk {
                background-color: #FFA756;
                border-radius: 10px;
            }
        """
    paused_stylesheet = ongoing_stylesheet + styles_to_overwride
    styles_to_overwride = f"""
            QProgressBar::chunk {{
                background-color: {GOGO_NORMAL_COLOR};
            }}"""
    completed_stylesheet = s = ongoing_stylesheet + styles_to_overwride
    styles_to_overwride = f"""
            QProgressBar::chunk {{
            background-color: {RED_NORMAL_COLOR}
            }}"""
    cancelled_stylesheet = ongoing_stylesheet + styles_to_overwride
    text_style_sheet = """
                    OutlinedLabel {
                    color: white;
                    font-size: 26px;
                    font-family: "Berlin Sans FB Demi";
                        }
                        """

    def __init__(
        self,
        parent: QWidget | None,
        task_title: str,
        item_task_is_applied_on: str,
        total_value: int,
        units: str,
        units_divisor: int = 1,
        delete_on_completion=True,
    ):
        super().__init__(parent)
        self.item_task_is_applied_on = item_task_is_applied_on
        self.total_value = total_value
        self.units = units
        self.units_divisor = units_divisor
        self.mutex = QMutex()
        self.items_layout = QHBoxLayout(self)
        self.delete_on_completion = delete_on_completion
        self.destructLater = lambda: QTimer(self).singleShot(40000, self.deleteLater)
        self.setLayout(self.items_layout)
        self.paused = False
        self.cancelled = False

        self.bar = QProgressBar(self)
        self.bar.setValue(0)
        self.bar.setMaximum(total_value)
        self.task_title = task_title
        self.bar.setFormat(f"{task_title} {item_task_is_applied_on}")
        self.bar.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self.bar.setStyleSheet(self.ongoing_stylesheet)
        height = 50
        self.percentage = OutlinedLabel(self, 1, 35)
        self.percentage.setText("0 %")
        self.percentage.setFixedHeight(height)
        self.percentage.setStyleSheet(self.text_style_sheet)

        self.rate = OutlinedLabel(self, 1, 40)
        self.rate.setText(f" 0 {units}/s")
        self.rate.setFixedHeight(height)
        self.rate.setStyleSheet(self.text_style_sheet)

        self.eta = OutlinedLabel(self, 1, 40)
        self.eta.setText("∞ secs left")
        self.eta.setFixedHeight(height)
        self.eta.setStyleSheet(self.text_style_sheet)

        self.current_against_max_values = OutlinedLabel(self, 1, 40)
        self.current_against_max_values.setText(
            f"0/{round(total_value/units_divisor)} {units}"
        )
        self.current_against_max_values.setFixedHeight(height)
        self.current_against_max_values.setStyleSheet(self.text_style_sheet)

        self.items_layout.addWidget(self.percentage)
        self.items_layout.addWidget(self.bar)
        self.items_layout.addWidget(self.eta)
        self.items_layout.addWidget(self.rate)
        self.items_layout.addWidget(self.current_against_max_values)
        self.prev_time = time()

    def is_complete(self) -> bool:
        return True if self.bar.value() >= self.total_value else False

    def cancel(self):
        if not self.paused and not self.is_complete() and not self.cancelled:
            self.bar.setFormat(f"Cancelled {self.item_task_is_applied_on}")
            self.cancelled = True
            self.bar.setStyleSheet(self.cancelled_stylesheet)
            if self.delete_on_completion:
                self.destructLater()

    def pause_or_resume(self):
        if not self.cancelled and not self.is_complete():
            self.paused = not self.paused
            if self.paused:
                self.bar.setStyleSheet(self.paused_stylesheet)
                self.bar.setFormat(f"Paused {self.item_task_is_applied_on}")
            else:
                self.bar.setStyleSheet(self.ongoing_stylesheet)
                self.bar.setFormat(f"{self.task_title} {self.item_task_is_applied_on}")

    def update_bar(self, added_value: int):
        self.mutex.lock()
        new_value = self.bar.value() + added_value
        curr_time = time()
        time_elapsed = curr_time - self.prev_time
        max_value = self.bar.maximum()
        complete = False
        if new_value >= max_value:
            new_value = max_value
            self.bar.setFormat(f"Completed {self.item_task_is_applied_on}")
            self.bar.setStyleSheet(self.completed_stylesheet)
            complete = True
        self.bar.setValue(new_value)
        percent_new_value = round(new_value / max_value * 100)
        self.percentage.setText(f"{percent_new_value}%")
        self.current_against_max_values.setText(
            f" {round(new_value/self.units_divisor)}/{round(max_value/self.units_divisor)} {self.units}"
        )
        # If statement to handle cases of annoying division by zero error where downloads update super quick so elapsed time is roughly zero
        if time_elapsed > 0 and added_value > 0:
            rate = added_value / time_elapsed
            eta = (max_value - new_value) * (1 / rate)
            if eta >= 3600:
                self.eta.setText(f"{round(eta/3600, 1)} hrs left")
            elif eta >= 60:
                self.eta.setText(f"{round(eta/60)} mins left")
            else:
                self.eta.setText(f"{round(eta)} secs left")
            self.rate.setText(f" {round(rate/self.units_divisor, 1)} {self.units}/s")
            self.prev_time = curr_time
        self.mutex.unlock()
        if complete and self.delete_on_completion:
            self.destructLater()


class ProgressBarWithButtons(ProgressBarWithoutButtons):
    def __init__(
        self,
        parent: QWidget | None,
        task_title: str,
        item_task_is_applied_on: str,
        total_value: int,
        units: str,
        units_divisor: int,
        pause_icon: Icon,
        resume_icon: Icon,
        cancel_icon: Icon,
        pause_callback: Callable[[], None],
        cancel_callback: Callable[[], None],
        delete_on_completion=True,
    ):
        super().__init__(
            parent,
            task_title,
            item_task_is_applied_on,
            total_value,
            units,
            units_divisor,
            delete_on_completion=delete_on_completion,
        )
        self.pause_icon = pause_icon
        self.resume_icon = resume_icon
        self.pause_callback = pause_callback
        self.cancel_callback = cancel_callback
        self.pause_button = IconButton(pause_icon, 1.3, self)
        self.cancel_button: IconButton = IconButton(cancel_icon, 1.3, self)
        self.pause_button.clicked.connect(self.pause_or_resume)
        self.cancel_button.clicked.connect(self.cancel)
        self.items_layout.addWidget(self.pause_button)
        self.items_layout.addWidget(self.cancel_button)

    def pause_or_resume(self):
        self.pause_callback()
        super().pause_or_resume()
        if self.paused:
            return self.pause_button.setIcon(self.resume_icon)
        self.pause_button.setIcon(self.pause_icon)

    def cancel(self):
        self.cancel_callback()
        super().cancel()


class ScrollableSection(QScrollArea):
    def __init__(self, layout: QHBoxLayout | QVBoxLayout):
        super().__init__()
        self.setWidgetResizable(True)
        self.main_widget = QWidget()
        self.main_widget.setLayout(layout)
        self.main_layout = self.layout()
        self.setWidget(self.main_widget)
        self.setStyleSheet(
            """
                    QWidget {
                        background-color: transparent;
                        border: None;
                        }"""
        )


class FolderButton(IconButton):
    def __init__(
        self, path: str, size_x: int, size_y: int, parent: QWidget | None = None
    ):
        super().__init__(Icon(size_x, size_y, FOLDER_ICON_PATH), 1.3, parent)
        self.folder_path = path
        self.clicked.connect(lambda: open_folder(self.folder_path))


class NumberInput(QLineEdit):
    def __init__(self, font_size: int = 14, parent: QWidget | None = None):
        super().__init__(parent)
        self.installEventFilter(self)
        self.setStyleSheet(
            f"""
            QLineEdit{{
                border: 2px solid black;
                border-radius: 12px;
                padding: 5px;
                color: black;
                font-size: {font_size}px;
                font-family: "Berlin Sans FB Demi";
                background-color: white;
            }}
        """
        )

    def eventFilter(self, a0: QObject | None, a1: QEvent | None):
        if a1 is not None and cast(QEvent, a1.type()) == QKeyEvent.Type.KeyPress:
            if cast(QKeyEvent, a1).key() in (
                Qt.Key.Key_0,
                Qt.Key.Key_1,
                Qt.Key.Key_2,
                Qt.Key.Key_3,
                Qt.Key.Key_4,
                Qt.Key.Key_5,
                Qt.Key.Key_6,
                Qt.Key.Key_7,
                Qt.Key.Key_8,
                Qt.Key.Key_9,
                Qt.Key.Key_Backspace,
                Qt.Key.Key_Delete,
                Qt.Key.Key_Left,
                Qt.Key.Key_Right,
            ):
                return False
            else:
                return True
        return super().eventFilter(a0, a1)


class QualityButton(OptionButton):
    def __init__(self, window: QWidget, quality: str, font_size: int):
        super().__init__(
            window, quality, quality, font_size, PAHE_NORMAL_COLOR, PAHE_PRESSED_COLOR
        )
        self.quality = quality


class SubDubButton(OptionButton):
    def __init__(self, window: QWidget, sub_or_dub: str, font_size: int):
        super().__init__(
            window,
            sub_or_dub,
            sub_or_dub.upper(),
            font_size,
            PAHE_NORMAL_COLOR,
            PAHE_PRESSED_COLOR,
        )
        self.sub_or_dub = sub_or_dub


class GogoNormOrHlsButton(OptionButton):
    def __init__(self, window: QWidget | None, norm_or_hls: str, font_size: int):
        super().__init__(
            window,
            norm_or_hls,
            norm_or_hls.upper(),
            font_size,
            RED_NORMAL_COLOR,
            RED_PRESSED_COLOR,
        )
        self.norm_or_hls = norm_or_hls
        if self.norm_or_hls == GOGO_NORM_MODE:
            return self.setToolTip(
                "Normal download functionality, similar to Animepahe but may occassionally fail"
            )
        self.setToolTip(
            "Guaranteed to work, it's like downloading a live stream as opposed to a file\nYou need to install FFmpeg for it to work but Senpwai will try to automatically install it"
        )


class HorizontalLine(QFrame):
    def __init__(self, color: str = "black", parent: QWidget | None = None):
        super().__init__(parent)
        self.setFrameShape(QFrame.Shape.HLine)
        self.setStyleSheet(
            f"""
                        QFrame {{ 
                            background-color: {color}; 
                            }}
                            """
        )


class Title(StyledLabel):
    def __init__(self, text: str, font=33):
        super().__init__(None, font, "orange", font_color="black")
        set_minimum_size_policy(self)
        self.setText(text)



from sys import platform as sysplatform
from typing import TYPE_CHECKING
from webbrowser import open_new_tab

from PyQt6.QtCore import Qt
from PyQt6.QtWidgets import QHBoxLayout, QVBoxLayout, QWidget
from senpwai.utils.static import (
    ABOUT_BCKG_IMAGE_PATH,
    DISCORD_ICON_PATH,
    GIGACHAD_AUDIO_PATH,
    GITHUB_ICON_PATH,
    GITHUB_REPO_URL,
    GITHUB_SPONSORS_ICON_PATH,
    HENTAI_ADDICT_AUDIO_PATH,
    HENTAI_ADDICT_ICON_PATH,
    MORBIUS_AUDIO_PATH,
    MORBIUS_IS_PEAK_ICON_PATH,
    PATREON_ICON_PATH,
    REDDIT_ICON_PATH,
    SEN_ICON_PATH,
    VERSION,
)
from senpwai.utils.widgets import (
    AudioPlayer,
    Icon,
    IconButton,
    ScrollableSection,
    StyledLabel,
    Title,
    set_minimum_size_policy,
)

from senpwai.windows.abstracts import AbstractWindow

# https://stackoverflow.com/questions/39740632/python-type-hinting-without-cyclic-imports/3957388#39757388
if TYPE_CHECKING:
    from senpwai.windows.main import MainWindow

class AboutWindow(AbstractWindow):
    def __init__(self, main_window: 'MainWindow'):
        super().__init__(main_window, ABOUT_BCKG_IMAGE_PATH)
        main_layout = QVBoxLayout()
        main_widget = ScrollableSection(main_layout)
        reviews_title = Title("Reviews")
        set_minimum_size_policy(reviews_title)
        reviews_widget = QWidget()
        reviews_layout = QHBoxLayout()
        size = 60
        sen_review = Review(
            Icon(size, size, SEN_ICON_PATH),
            GIGACHAD_AUDIO_PATH,
            "SenZmaKi",
            "69/10 Truly one of the apps of all time.",
        )
        hentai_addict_review = Review(
            Icon(size, size, HENTAI_ADDICT_ICON_PATH),
            HENTAI_ADDICT_AUDIO_PATH,
            "HentaiAddict01",
            "0/10 Can't even batch download hentai.. .",
        )
        morbius_is_peak_review = Review(
            Icon(size, size, MORBIUS_IS_PEAK_ICON_PATH),
            MORBIUS_AUDIO_PATH,
            "MorbiusIsPeak",
            "4/10 Morbius better + ratio + morbiusless",
        )
        reviews_layout.addWidget(sen_review)
        reviews_layout.addWidget(hentai_addict_review)
        reviews_layout.addWidget(morbius_is_peak_review)
        reviews_widget.setLayout(reviews_layout)
        set_minimum_size_policy(reviews_widget)
        support_title = Title("Support")
        set_minimum_size_policy(support_title)
        donations_label = StyledLabel()
        donations_label.setText(
            "Konnichiwa 👋🏿 it's Sen, the creator of Senpwai, I'm a goofy ahh ahh college student from Kenya.\nWifi is kinda expensive in my country, so donations help me pay for internet to keep developing new features and fixing bugs"
        )
        set_minimum_size_policy(donations_label)
        github_sponsors_button = IconButton(
            Icon(120, 120, GITHUB_SPONSORS_ICON_PATH), 1.1
        )
        github_sponsors_button.clicked.connect(
            lambda: open_new_tab("https://github.com/sponsors/SenZmaKi")
        )
        github_sponsors_button.setToolTip("https://github.com/sponsors/SenZmaKi")
        patreon_button = IconButton(Icon(80, 80, PATREON_ICON_PATH), 1.1)
        patreon_button.clicked.connect(
            lambda: open_new_tab("https://patreon.com/Senpwai")
        )
        patreon_button.setToolTip("https://patreon.com/Senpwai")
        donation_buttons_widget = QWidget()
        donation_buttons_layout = QHBoxLayout()
        donation_buttons_layout.addWidget(github_sponsors_button)
        donation_buttons_layout.addSpacing(10)
        donation_buttons_layout.addWidget(patreon_button)
        donation_buttons_layout.setAlignment(Qt.AlignmentFlag.AlignLeft)
        donation_buttons_widget.setLayout(donation_buttons_layout)
        leave_a_star_label = StyledLabel()
        leave_a_star_label.setText(
            "You can also support Senpwai by leaving a star on the github repo, stars help other weebs like us know about it"
        )
        set_minimum_size_policy(leave_a_star_label)
        social_links_title = Title("Social Links")
        set_minimum_size_policy(social_links_title)
        bug_reports_label = StyledLabel()
        bug_reports_label.setText(
            "Found a bug or wanna suggest a feature? Report/Suggest it in the discord server, github issues or subreddit"
        )
        set_minimum_size_policy(bug_reports_label)
        github_button = IconButton(Icon(200, 80, GITHUB_ICON_PATH), 1.1)
        github_button.clicked.connect(lambda: open_new_tab(GITHUB_REPO_URL))
        github_button.setToolTip(GITHUB_REPO_URL)
        reddit_button = IconButton(Icon(80, 80, REDDIT_ICON_PATH), 1.1)
        reddit_button.clicked.connect(
            lambda: open_new_tab("https://reddit.com/r/Senpwai")
        )
        reddit_button.setToolTip("https://reddit.com/r/Senpwai")
        discord_button = IconButton(Icon(80, 80, DISCORD_ICON_PATH), 1.1)
        discord_button.clicked.connect(
            lambda: open_new_tab("https://discord.gg/e9UxkuyDX2")
        )
        discord_button.setToolTip("https://discord.gg/e9UxkuyDX2")
        social_links_buttons_widget = QWidget()
        social_links_buttons_layout = QHBoxLayout()
        social_links_buttons_layout.addWidget(github_button)
        social_links_buttons_layout.addSpacing(30)
        social_links_buttons_layout.addWidget(reddit_button)
        social_links_buttons_layout.addSpacing(30)
        social_links_buttons_layout.addWidget(discord_button)
        social_links_buttons_layout.setAlignment(Qt.AlignmentFlag.AlignLeft)
        social_links_buttons_widget.setLayout(social_links_buttons_layout)

        uninstall_title = Title("Uninstall Info")
        set_minimum_size_policy(uninstall_title)
        uninstall_info_label = StyledLabel()
        uninstall_info_label.setText(
            'To completely remove Senpwai (don\'t know why you would though), post-uninstallation press "Win + R",\ntype "%appdata%\\..\\Local\\Programs" and press enter. Look for a folder named "Senpwai" then delete it'
        )
        set_minimum_size_policy(uninstall_info_label)

        version_title = Title(f"Version {VERSION}")
        set_minimum_size_policy(version_title)

        main_layout.addSpacing(40)
        main_layout.addWidget(reviews_title, Qt.AlignmentFlag.AlignCenter)
        main_layout.addWidget(reviews_widget, Qt.AlignmentFlag.AlignCenter)
        main_layout.addWidget(support_title)
        main_layout.addWidget(donations_label)
        main_layout.addWidget(donation_buttons_widget)
        main_layout.addWidget(leave_a_star_label)
        main_layout.addSpacing(40)
        main_layout.addWidget(social_links_title)
        main_layout.addWidget(bug_reports_label)
        main_layout.addWidget(social_links_buttons_widget)
        if sysplatform == "win32":
            main_layout.addWidget(uninstall_title)
            main_layout.addWidget(uninstall_info_label)
        main_layout.addWidget(version_title)
        self.full_layout.addWidget(main_widget, Qt.AlignmentFlag.AlignTop)
        self.setLayout(self.full_layout)


class Review(QWidget):
    def __init__(self, icon: Icon, audio_path: str, author: str, text: str):
        super().__init__()
        main_layout = QVBoxLayout()

        profile_pic = IconButton(icon, 1.1)
        profile_pic.clicked.connect(AudioPlayer(self, audio_path, volume=60).play)
        author_name = StyledLabel(None, 15, "orange", font_color="black")
        author_name.setText(author)
        set_minimum_size_policy(author_name)
        review_text = StyledLabel(None, 15, font_color="white", bckg_color="black")
        review_text.setText(text)
        set_minimum_size_policy(review_text)

        main_layout.addWidget(profile_pic, alignment=Qt.AlignmentFlag.AlignHCenter)
        main_layout.addWidget(author_name, alignment=Qt.AlignmentFlag.AlignHCenter)
        main_layout.addWidget(review_text)
        self.setLayout(main_layout)
        set_minimum_size_policy(self)



from typing import TYPE_CHECKING, cast

from PyQt6.QtCore import QSize, Qt, QThread, QTimer, pyqtSignal
from PyQt6.QtGui import QPixmap
from PyQt6.QtWidgets import (
    QHBoxLayout,
    QLabel,
    QScrollBar,
    QSpacerItem,
    QVBoxLayout,
    QWidget,
)
from senpwai.utils.classes import SETTINGS, Anime, AnimeDetails
from senpwai.utils.scraper import lacked_episode_numbers
from senpwai.utils.static import (
    CHOSEN_ANIME_WINDOW_BCKG_IMAGE_PATH,
    DUB,
    GOGO,
    GOGO_HOVER_COLOR,
    GOGO_NORMAL_COLOR,
    Q_360,
    Q_480,
    Q_720,
    Q_1080,
    RED_NORMAL_COLOR,
    RED_PRESSED_COLOR,
    SUB,
)
from senpwai.utils.widgets import (
    DualStateButton,
    ErrorLabel,
    FolderButton,
    GogoNormOrHlsButton,
    HorizontalLine,
    NumberInput,
    QualityButton,
    ScrollableSection,
    StyledButton,
    StyledLabel,
    SubDubButton,
    set_minimum_size_policy,
)
import os

from senpwai.windows.download import DownloadWindow
from senpwai.windows.settings import SettingsWindow
from senpwai.windows.abstracts import AbstractTemporaryWindow
# https://stackoverflow.com/questions/39740632/python-type-hinting-without-cyclic-imports/3957388#39757388
if TYPE_CHECKING:
    from senpwai.windows.main import MainWindow

class SummaryLabel(StyledLabel):
    def __init__(self, summary: str):
        super().__init__(font_size=18)
        self.setText(summary)
        self.setWordWrap(True)


class HavedEpisodes(StyledLabel):
    def __init__(
        self,
        start: int | None,
        end: int | None,
        haved_count: int | None,
        total_episode_count: int,
    ):
        super().__init__(font_size=18)
        self.start = start
        self.end = end
        self.count = haved_count
        if not haved_count:
            self.setText("You have No episodes of this anime.")
        elif haved_count >= total_episode_count:
            self.setText("You have all the current episodes of this anime, weeeeb.")
        else:
            self.setText(
                f"You have {haved_count} episodes from {start} to {end}."
            ) if haved_count != 1 else self.setText(
                f"You have {haved_count} episode from {start} to {end}."
            )


class MakeAnimeDetailsThread(QThread):
    finished = pyqtSignal(AnimeDetails)

    def __init__(self, window: 'MainWindow', anime: Anime, site: str):
        super().__init__(window)
        self.anime = anime
        self.site = site
        self.window = window

    def run(self):
        self.finished.emit(AnimeDetails(self.anime, self.site))


class ChosenAnimeWindow(AbstractTemporaryWindow):
    def __init__(self, main_window: 'MainWindow', anime_details: AnimeDetails):
        super().__init__(main_window, CHOSEN_ANIME_WINDOW_BCKG_IMAGE_PATH)
        self.main_window = main_window
        self.anime_details = anime_details

        main_layout = QHBoxLayout()
        left_widgets_widget = QWidget()
        left_widgets_layout = QVBoxLayout()
        poster = Poster(self.anime_details.metadata.get_poster_bytes())
        left_widgets_layout.addWidget(poster)
        left_widgets_widget.setLayout(left_widgets_layout)
        bottom_widgets = QWidget()
        bottom_layout = QVBoxLayout()
        bottom_widgets.setLayout(bottom_layout)
        bottom_top_widget = QWidget()
        bottom_top_layout = QHBoxLayout()
        bottom_top_widget.setLayout(bottom_top_layout)
        release_year = StyledLabel(None, 21, "black")
        release_year.setText(str(anime_details.metadata.release_year))
        set_minimum_size_policy(release_year)
        bottom_top_layout.addWidget(release_year)
        airing_status = StyledLabel(None, 21, "blue")
        airing_status.setText(anime_details.metadata.airing_status)
        set_minimum_size_policy(airing_status)
        bottom_top_layout.addWidget(airing_status)
        self.episode_count = EpisodeCount(str(self.anime_details.episode_count))
        set_minimum_size_policy(self.episode_count)
        bottom_top_layout.addWidget(self.episode_count)
        bottom_bottom_widget = QWidget()
        bottom_bottom_layout = QHBoxLayout()
        bottom_bottom_widget.setLayout(bottom_bottom_layout)
        for genre in anime_details.metadata.genres[:3]:
            g_wid = StyledLabel(None, 21, "orange")
            g_wid.setText(genre)
            set_minimum_size_policy(g_wid)
            bottom_bottom_layout.addWidget(g_wid)
        left_widgets_layout.addWidget(bottom_top_widget)
        left_widgets_layout.addWidget(bottom_bottom_widget)
        set_minimum_size_policy(left_widgets_widget)
        main_layout.addWidget(left_widgets_widget)
        right_widgets_widget = QWidget()
        right_widgets_layout = QVBoxLayout()
        title = Title(self.anime_details.anime.title)
        right_widgets_layout.addWidget(title)
        line_under_title = HorizontalLine(parent=self)
        line_under_title.setFixedHeight(10)
        right_widgets_layout.addWidget(line_under_title)
        summary_label = SummaryLabel(self.anime_details.metadata.summary)
        summary_layout = QVBoxLayout()
        summary_layout.addWidget(summary_label)
        summary_widget = ScrollableSection(summary_layout)
        summary_widget.setMaximumHeight(300)
        right_widgets_layout.addWidget(summary_widget)

        self.sub_button = SubDubButton(self, SUB, 18)
        set_minimum_size_policy(self.sub_button)
        self.dub_button = None
        self.sub_button.clicked.connect(lambda: self.update_sub_or_dub(SUB))
        if SETTINGS.sub_or_dub == SUB:
            self.sub_button.set_picked_status(True)
        if self.anime_details.dub_available:
            self.dub_button = SubDubButton(self, DUB, 18)
            set_minimum_size_policy(self.dub_button)
            self.dub_button.clicked.connect(lambda: self.update_sub_or_dub(DUB))
            if SETTINGS.sub_or_dub == DUB:
                self.dub_button.set_picked_status(True)
            self.dub_button.clicked.connect(
                lambda: self.sub_button.set_picked_status(False)
            )
            self.sub_button.clicked.connect(
                lambda: cast(SubDubButton, self.dub_button).set_picked_status(False)
            )
        else:
            self.sub_button.set_picked_status(True)
            self.anime_details.sub_or_dub = SUB
        first_row_of_buttons_widget = QWidget()
        first_row_of_buttons_layout = QHBoxLayout()
        first_row_of_buttons_layout.addWidget(self.sub_button)
        if self.dub_button:
            first_row_of_buttons_layout.addWidget(self.dub_button)
        first_row_of_buttons_layout.addSpacerItem(QSpacerItem(20, 0))
        self.button_1080 = QualityButton(self, Q_1080, 18)
        self.button_720 = QualityButton(self, Q_720, 18)
        self.button_480 = QualityButton(self, Q_480, 18)
        self.button_360 = QualityButton(self, Q_360, 18)
        self.quality_buttons_list = [
            self.button_1080,
            self.button_720,
            self.button_480,
            self.button_360,
        ]

        for button in self.quality_buttons_list:
            set_minimum_size_policy(button)
            first_row_of_buttons_layout.addWidget(button)
            quality = button.quality
            button.clicked.connect(
                lambda garbage_bool, quality=quality: self.update_quality(quality)
            )
            if quality == SETTINGS.quality:
                button.set_picked_status(True)
        first_row_of_buttons_layout.addSpacerItem(QSpacerItem(20, 0))
        if anime_details.site == GOGO:
            self.norm_button = GogoNormOrHlsButton(self, "norm", 18)
            self.hls_button = GogoNormOrHlsButton(self, "hls", 18)
            self.norm_button.clicked.connect(
                lambda garbage_bool: self.update_is_hls_download(False)
            )
            self.hls_button.clicked.connect(
                lambda garbage_bool: self.update_is_hls_download(True)
            )
            set_minimum_size_policy(self.norm_button)
            set_minimum_size_policy(self.hls_button)
            if anime_details.is_hls_download:
                self.hls_button.set_picked_status(True)
            else:
                self.norm_button.set_picked_status(True)
            first_row_of_buttons_layout.addWidget(self.norm_button)
            first_row_of_buttons_layout.addWidget(self.hls_button)

        first_row_of_buttons_layout.setAlignment(Qt.AlignmentFlag.AlignLeft)
        first_row_of_buttons_widget.setLayout(first_row_of_buttons_layout)
        right_widgets_layout.addWidget(first_row_of_buttons_widget)

        second_row_of_buttons_widget = QWidget()
        second_row_of_buttons_layout = QHBoxLayout()

        start_episode = (
            str((self.anime_details.haved_end) + 1)
            if (
                self.anime_details.haved_end
                and self.anime_details.haved_end < self.anime_details.episode_count
            )
            else "1"
        )
        input_size = QSize(80, 40)
        if anime_details.metadata.airing_status != "UPCOMING":
            self.start_episode_input = NumberInput(21)
            self.start_episode_input.setFixedSize(input_size)
            self.start_episode_input.setPlaceholderText("START")
            self.start_episode_input.setText(str(start_episode))
            self.end_episode_input = NumberInput(21)
            self.end_episode_input.setPlaceholderText("STOP")
            self.end_episode_input.setFixedSize(input_size)
            self.download_button = DownloadButton(
                self, self.main_window.download_window, self.anime_details
            )
            set_minimum_size_policy(self.download_button)
            second_row_of_buttons_layout.addWidget(self.start_episode_input)
            second_row_of_buttons_layout.addWidget(self.end_episode_input)
            second_row_of_buttons_layout.addWidget(self.download_button)
            QTimer.singleShot(0, self.download_button.setFocus)
        including_error_label_widget = QWidget()
        including_error_label_layout = QVBoxLayout()
        self.error_label = ErrorLabel(18, 6)
        self.error_label.hide()
        including_error_label_layout.addWidget(self.error_label)
        second_row_of_buttons_widget.setLayout(second_row_of_buttons_layout)
        second_row_of_buttons_layout.setAlignment(Qt.AlignmentFlag.AlignLeft)
        including_error_label_layout.addWidget(second_row_of_buttons_widget)
        including_error_label_widget.setLayout(including_error_label_layout)
        right_widgets_layout.addWidget(including_error_label_widget)

        third_row_of_labels_widget = QWidget()
        third_row_of_labels_layout = QHBoxLayout()

        haved_episodes = HavedEpisodes(
            self.anime_details.haved_start,
            self.anime_details.haved_end,
            self.anime_details.haved_count,
            self.anime_details.episode_count,
        )
        set_minimum_size_policy(haved_episodes)
        third_row_of_labels_layout.addWidget(haved_episodes)
        if os.path.isdir(self.anime_details.anime_folder_path):
            folder_button = FolderButton(self.anime_details.anime_folder_path, 100, 100)
            third_row_of_labels_layout.addSpacerItem(QSpacerItem(20, 0))
            third_row_of_labels_layout.addWidget(folder_button)
        track_button = TrackButton(
            anime_details.anime.title, self, self.main_window.settings_window
        )
        if anime_details.anime.title in SETTINGS.tracked_anime:
            track_button.change_status()
        set_minimum_size_policy(track_button)
        third_row_of_labels_layout.addSpacerItem(QSpacerItem(20, 0))
        third_row_of_labels_layout.addWidget(track_button)
        third_row_of_labels_layout.setAlignment(Qt.AlignmentFlag.AlignLeft)
        third_row_of_labels_widget.setLayout(third_row_of_labels_layout)
        right_widgets_layout.addWidget(third_row_of_labels_widget)
        right_widgets_widget.setLayout(right_widgets_layout)
        main_layout.addWidget(right_widgets_widget)
        main_widget = ScrollableSection(main_layout)
        bar = cast(QScrollBar, main_widget.horizontalScrollBar())
        bar.setValue(bar.maximum())
        self.full_layout.addWidget(main_widget)
        self.setLayout(self.full_layout)
        # For testing purposes
        # self.download_button.animateClick()

    def update_quality(self, quality: str):
        self.anime_details.quality = quality
        for button in self.quality_buttons_list:
            if button.quality != quality:
                button.set_picked_status(False)

    def update_is_hls_download(self, is_hls_download: bool):
        self.anime_details.is_hls_download = is_hls_download
        if is_hls_download:
            self.norm_button.set_picked_status(False)
        else:
            self.hls_button.set_picked_status(False)

    def update_sub_or_dub(self, sub_or_dub: str):
        self.anime_details.sub_or_dub = sub_or_dub
        if sub_or_dub == DUB:
            self.sub_button.set_picked_status(False)
        elif self.dub_button is not None:
            self.dub_button.set_picked_status(False)

    def error(self, error_message: str):
        self.error_label.setText(error_message)
        self.error_label.update()
        set_minimum_size_policy(self.error_label)
        self.error_label.show()


class TrackButton(DualStateButton):
    def __init__(
        self,
        anime_title: str,
        chosen_anime_window: ChosenAnimeWindow,
        settings: SettingsWindow,
    ):
        self.anime_title = anime_title
        self.off_tooltip = f"Add {self.anime_title} to your tracked anime list\nfor automatic downloading of new episodes"
        self.add_to_settings = lambda: settings.tracked_anime.add_anime(anime_title)
        self.remove_from_settings = lambda: settings.tracked_anime.remove_anime(
            anime_title
        )
        self.set = False
        super().__init__(
            chosen_anime_window,
            18,
            "white",
            RED_PRESSED_COLOR,
            RED_NORMAL_COLOR,
            "TRACK",
            "UNTRACK",
        )
        self.setToolTip(self.off_tooltip)

    def change_status(self):
        super().change_status()
        if self.on:
            self.setToolTip(f"Remove {self.anime_title} from your tracked anime list")
            return self.add_to_settings()
        self.remove_from_settings()
        self.setToolTip(self.off_tooltip)


class DownloadButton(StyledButton):
    def __init__(
        self,
        chosen_anime_window: ChosenAnimeWindow,
        download_window: DownloadWindow,
        anime_details: AnimeDetails,
    ):
        super().__init__(
            chosen_anime_window,
            21,
            "white",
            "green",
            GOGO_NORMAL_COLOR,
            GOGO_HOVER_COLOR,
        )
        self.chosen_anime_window = chosen_anime_window
        self.download_window = download_window
        self.main_window = chosen_anime_window.main_window
        self.anime_details = anime_details
        self.clicked.connect(self.handle_download_button_clicked)
        self.setText("DOWNLOAD")

    def handle_download_button_clicked(self):
        invalid_input = False
        episode_count = self.anime_details.episode_count
        start_episode = self.chosen_anime_window.start_episode_input.text()
        end_episode = self.chosen_anime_window.end_episode_input.text()
        predicted_start_point = self.anime_details.haved_end
        predicted_start_point = (
            predicted_start_point + 1 if predicted_start_point is not None else 1
        )
        error = self.chosen_anime_window.error
        if episode_count == 0:
            error("This anime has no episodes yet on this site.")
            invalid_input = True
        if "0" == start_episode or "0" == end_episode:
            error("What am I supposed to do with a zero?")
            invalid_input = True
        if not start_episode:
            start_episode = 1
            self.chosen_anime_window.start_episode_input.setText("1")
        if not end_episode:
            end_episode = episode_count
            self.chosen_anime_window.end_episode_input.setText("")
        start_episode = int(start_episode)
        end_episode = int(end_episode)

        if (
            predicted_start_point <= episode_count
            and start_episode > episode_count
            and not invalid_input
        ):
            error(
                "Start episode cannot be greater than the number of episodes the anime has."
            )
            invalid_input = True
            start_episode = predicted_start_point
            self.chosen_anime_window.start_episode_input.setText(
                str(predicted_start_point)
            )

        if (end_episode < start_episode) and not invalid_input:
            error("Stop episode cannot be less than start episode, hontoni baka ga.")
            invalid_input = True
            end_episode = episode_count
            self.chosen_anime_window.end_episode_input.setText("")

        elif (end_episode > episode_count) and not invalid_input:
            error(
                "Oe oe oe mate, stop episode cannot be greater than the number of episodes this anime has."
            )
            end_episode = episode_count
            self.chosen_anime_window.end_episode_input.setText("")
            invalid_input = True

        if not invalid_input:
            self.anime_details.lacked_episode_numbers = lacked_episode_numbers(
                start_episode, end_episode, self.anime_details.haved_episodes
            )
            if len(self.anime_details.lacked_episode_numbers) == 0:
                error(
                    "Bakayorou, you already have all episodes within the provided range!!!"
                )
                invalid_input = True

        if invalid_input:
            return self.chosen_anime_window.episode_count.brighten()
        self.main_window.stacked_windows.setCurrentWidget(
            self.main_window.download_window
        )
        self.download_window.initiate_download_pipeline(self.anime_details)
        self.main_window.stacked_windows.removeWidget(self.chosen_anime_window)
        self.chosen_anime_window.deleteLater()


class EpisodeCount(StyledLabel):
    def __init__(self, count: str):
        super().__init__(None, 21, "rgba(255, 50, 0, 250)")
        self.setText(f"{count} episodes")
        self.normal_style_sheet = self.styleSheet()
        self.setWordWrap(True)
        self.bright_stylesheet = (
            self.normal_style_sheet
            + """
            QLabel {
                color: black;
                background-color: rgba(255, 0, 0, 255);
                border: 2px solid black;
                    }
                    """
        )

    def brighten(self):
        self.setStyleSheet(self.bright_stylesheet)

        def revert_styles():
            return self.setStyleSheet(self.normal_style_sheet)

        timer = QTimer(self)
        timer.timeout.connect(revert_styles)
        timer.start(6 * 1000)


class Poster(QLabel):
    def __init__(self, image: bytes):
        super().__init__()
        size_x = 350
        size_y = 500
        pixmap = QPixmap()
        pixmap.loadFromData(image)
        pixmap = pixmap.scaled(size_x, size_y, Qt.AspectRatioMode.IgnoreAspectRatio)

        self.move(50, 60)
        self.setPixmap(pixmap)
        self.setFixedSize(size_x, size_y)
        self.setStyleSheet(
            """
                        QLabel {
                        background-color: rgba(255, 160, 0, 255);
                        border-radius: 10px;
                        padding: 5px;
                        border: 2px solid black;
                        }
                        """
        )


class Title(StyledLabel):
    def __init__(self, title: str):
        super().__init__(None, 40, "orange", 20, "black")
        self.setText(title.upper())
        self.setWordWrap(True)



import os
import subprocess
import sys
from typing import TYPE_CHECKING, Any, Callable
from webbrowser import open_new_tab

from PyQt6.QtCore import Qt, QThread, pyqtSignal
from PyQt6.QtWidgets import QHBoxLayout, QVBoxLayout, QWidget
from senpwai.utils.classes import AnimeDetails, update_available
from senpwai.utils.scraper import (
    CLIENT,
    IBYTES_TO_MBS_DIVISOR,
    RESOURCE_MOVED_STATUS_CODES,
    Download,
    try_installing_ffmpeg,
)
from senpwai.utils.static import (
    APP_NAME,
    CANCEL_ICON_PATH,
    CHOPPER_CRYING_PATH,
    GITHUB_API_LATEST_RELEASE_ENDPOINT,
    GITHUB_ICON_PATH,
    GITHUB_REPO_URL,
    GOGO_HOVER_COLOR,
    GOGO_NORMAL_COLOR,
    GOGO_PRESSED_COLOR,
    IS_EXECUTABLE,
    IS_PIP_INSTALL,
    PAUSE_ICON_PATH,
    RED_HOVER_COLOR,
    RED_NORMAL_COLOR,
    RED_PRESSED_COLOR,
    RESUME_ICON_PATH,
    ROOT_DIRECTORY,
    UPDATE_BCKG_IMAGE_PATH,
    VERSION,
)
from senpwai.utils.widgets import (
    Icon,
    IconButton,
    StyledButton,
    StyledLabel,
    StyledTextBrowser,
    Title,
    set_minimum_size_policy,
)

from senpwai.windows.download import ProgressBarWithButtons
from senpwai.windows.abstracts import (
    AbstractTemporaryWindow,
    AbstractWindow,
)
# https://stackoverflow.com/questions/39740632/python-type-hinting-without-cyclic-imports/3957388#39757388
if TYPE_CHECKING:
    from senpwai.windows.main import MainWindow


class MiscWindow(AbstractTemporaryWindow):
    def __init__(self, main_window: 'MainWindow', misc_info_text: str):
        super().__init__(main_window, CHOPPER_CRYING_PATH)
        self.info_label = StyledLabel(font_size=25)
        self.info_label.setText(misc_info_text)
        set_minimum_size_policy(self.info_label)
        self.main_layout = QVBoxLayout()
        self.main_layout.addWidget(
            self.info_label, alignment=Qt.AlignmentFlag.AlignCenter
        )
        self.buttons_layout = QHBoxLayout()
        buttons_widget = QWidget()
        buttons_widget.setLayout(self.buttons_layout)
        self.main_layout.addWidget(buttons_widget)
        main_widget = QWidget()
        main_widget.setLayout(self.main_layout)
        self.full_layout.addWidget(main_widget, Qt.AlignmentFlag.AlignHCenter)
        self.setLayout(self.full_layout)


class NewVersionInfoWindow(MiscWindow):
    def __init__(self, main_window: 'MainWindow', info_text: str):
        super().__init__(main_window, info_text)
        title = Title(f"Version {VERSION} Changes")
        set_minimum_size_policy(title)
        self.main_layout.insertWidget(0, title, alignment=Qt.AlignmentFlag.AlignHCenter)


class NoFFmpegWindow(MiscWindow):
    initiate_download_pipeline = pyqtSignal(AnimeDetails)

    def __init__(self, main_window: 'MainWindow', anime_details: AnimeDetails):
        self.main_window = main_window
        info_text = "Sumanai, in order to use HLS mode you need to have\nFFmpeg installed and properly added to path"
        super().__init__(main_window, info_text)
        super().__init__(main_window, info_text)
        install_ffmepg_button = StyledButton(
            None, 25, "black", GOGO_NORMAL_COLOR, GOGO_HOVER_COLOR, GOGO_PRESSED_COLOR
        )
        install_ffmepg_button.setText("Automatically Install FFmpeg")
        set_minimum_size_policy(install_ffmepg_button)
        install_ffmepg_button.clicked.connect(
            TryInstallingFFmpegThread(self, anime_details).start
        )
        anime_details.is_hls_download = False
        download_in_normal_mode = StyledButton(
            None, 25, "black", RED_NORMAL_COLOR, RED_HOVER_COLOR, RED_PRESSED_COLOR
        )
        self.initiate_download_pipeline.connect(
            main_window.download_window.initiate_download_pipeline
        )
        download_in_normal_mode.clicked.connect(
            lambda: self.initiate_download_pipeline.emit(anime_details)
        )
        download_in_normal_mode.clicked.connect(self.download_window_button.click)
        download_in_normal_mode.setText("Download in Normal mode")
        set_minimum_size_policy(download_in_normal_mode)
        self.buttons_layout.addWidget(download_in_normal_mode)
        self.buttons_layout.addWidget(install_ffmepg_button)


class TryInstallingFFmpegThread(QThread):
    initiate_download_pipeline = pyqtSignal(AnimeDetails)
    switch_to_download_window = pyqtSignal()

    def __init__(self, no_ffmpeg_window: NoFFmpegWindow, anime_details: AnimeDetails):
        super().__init__(no_ffmpeg_window)
        self.initiate_download_pipeline.connect(
            lambda: no_ffmpeg_window.main_window.download_window.initiate_download_pipeline(
                anime_details
            )
        )
        self.anime_details = anime_details
        self.switch_to_download_window.connect(
            no_ffmpeg_window.download_window_button.click
        )
        self.no_ffmpeg_window = no_ffmpeg_window

    def start_download(self):
        self.anime_details.is_hls_download = True
        self.initiate_download_pipeline.emit(self.anime_details)
        self.switch_to_download_window.emit()

    def run(self):
        if try_installing_ffmpeg():
            self.no_ffmpeg_window.main_window.tray_icon.make_notification(
                "Successfully Installed", "FFmpeg", True, None
            )
            self.start_download()
        else:
            self.no_ffmpeg_window.main_window.tray_icon.make_notification(
                "Installation Failed",
                "Failed to automatically install FFmpeg, opened a guide on your browser uWu",
                False,
            )


class UpdateWindow(AbstractWindow):
    def __init__(
        self,
        main_window: 'MainWindow',
        download_url: str,
        file_name: str,
        update_info: str,
    ):
        super().__init__(main_window, UPDATE_BCKG_IMAGE_PATH)
        main_widget = QWidget()
        main_layout = QVBoxLayout()
        self.progress_bar: ProgressBarWithButtons | None = None
        before_click_label = StyledLabel(font_size=20)
        before_click_label.setWordWrap(True)
        update_info_text_browser = StyledTextBrowser(font_size=20)
        update_info_text_browser.setMarkdown(update_info)
        update_info_text_browser.setMinimumWidth(500)
        update_info_text_browser.setMinimumHeight(300)

        main_layout.addWidget(
            update_info_text_browser, alignment=Qt.AlignmentFlag.AlignCenter
        )
        if IS_PIP_INSTALL:
            before_click_label.setText(
                'Run "pip install senpwai --upgrade" to update to the latest version'
            )
            main_layout.addWidget(before_click_label, alignment=Qt.AlignmentFlag.AlignCenter)
        elif sys.platform == "win32" and IS_EXECUTABLE:
            before_click_label.setText(
                "Before you click the update button, ensure you don't have any active downloads cause Senpwai will restart"
            )
            main_layout.addWidget(
                before_click_label, alignment=Qt.AlignmentFlag.AlignCenter
            )
            self.update_button = StyledButton(
                self,
                24,
                "black",
                RED_NORMAL_COLOR,
                RED_HOVER_COLOR,
                RED_PRESSED_COLOR,
                20,
            )
            self.update_button.setText("DOWNLOAD AND INSTALL UPDATE")
            set_minimum_size_policy(self.update_button)
            download_widget = QWidget()
            self.download_layout = QVBoxLayout()
            self.download_layout.addWidget(
                self.update_button, alignment=Qt.AlignmentFlag.AlignCenter
            )
            main_layout.addWidget(download_widget)
            download_widget.setLayout(self.download_layout)
            self.update_button.clicked.connect(
                DownloadUpdateThread(
                    main_window, self, download_url, file_name
                ).start
            )
        else:
            before_click_label.setText(
                "\nBuild from source to update to the latest version, there is a guide on the README.md in the repository"
            )
            main_layout.addWidget(
                before_click_label, alignment=Qt.AlignmentFlag.AlignCenter
            )
            github_button = IconButton(Icon(300, 100, GITHUB_ICON_PATH), 1.1)
            github_button.clicked.connect(lambda: open_new_tab(GITHUB_REPO_URL))
            github_button.setToolTip(GITHUB_REPO_URL)
            main_layout.addWidget(
                github_button, alignment=Qt.AlignmentFlag.AlignCenter
            )

        set_minimum_size_policy(before_click_label)
        main_widget.setLayout(main_layout)
        main_layout.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self.full_layout.addWidget(main_widget, Qt.AlignmentFlag.AlignHCenter)
        self.setLayout(self.full_layout)

    def receive_total_size(self, size: int):
        pause_icon = Icon(40, 40, PAUSE_ICON_PATH)
        resume_icon = Icon(40, 40, RESUME_ICON_PATH)
        cancel_icon = Icon(40, 40, CANCEL_ICON_PATH)
        self.progress_bar = ProgressBarWithButtons(
            self,
            "Downloading",
            "update",
            size,
            "MB",
            IBYTES_TO_MBS_DIVISOR,
            pause_icon,
            resume_icon,
            cancel_icon,
            lambda: None,
            lambda: None,
        )
        self.download_layout.addWidget(self.progress_bar)
        self.update_button.hide()
        self.update_button.deleteLater()


class DownloadUpdateThread(QThread):
    update_bar = pyqtSignal(int)
    total_size = pyqtSignal(int)
    quit_app = pyqtSignal()

    def __init__(
        self,
        main_window: 'MainWindow',
        update_window: UpdateWindow,
        download_url: str,
        file_name: str,
    ):
        super().__init__(main_window)
        self.quit_app.connect(main_window.app.quit)
        self.download_url = download_url
        self.total_size.connect(update_window.receive_total_size)
        self.update_window = update_window
        self.file_name = file_name

    def run(self):
        response = CLIENT.get(self.download_url, stream=True)
        if response.status_code in RESOURCE_MOVED_STATUS_CODES:
            self.download_url = response.headers["Location"]
            response = CLIENT.get(self.download_url, stream=True)
        total_size = int(response.headers["Content-Length"])
        self.total_size.emit(total_size)
        self.update_window.progress_bar
        while not self.update_window.progress_bar:
            continue
        self.update_bar.connect(self.update_window.progress_bar.update_bar)
        file_name_no_ext, ext = os.path.splitext(self.file_name)
        download = Download(
            self.download_url,
            file_name_no_ext,
            ROOT_DIRECTORY,
            self.update_bar.emit,
            ext,
        )
        self.update_window.progress_bar.pause_callback = download.pause_or_resume
        self.update_window.progress_bar.cancel_callback = download.cancel
        download.start_download()
        if not download.cancelled:
            subprocess.Popen(
                [os.path.join(ROOT_DIRECTORY, self.file_name), "/silent", "/update"]
            )
            self.quit_app.emit()


class CheckIfUpdateAvailableThread(QThread):
    finished = pyqtSignal(tuple)

    def __init__(
        self,
        main_window: 'MainWindow',
        finished_callback: Callable[[tuple[bool, str, str, str]], Any],
    ):
        super().__init__(main_window)
        self.finished.connect(finished_callback)

    def run(self):
        self.finished.emit(
            update_available(GITHUB_API_LATEST_RELEASE_ENDPOINT, APP_NAME, VERSION)
        )

